import os
from tempfile import TemporaryDirectory
import pytest
from nabu.pipeline.config import (
    generate_nabu_configfile,
    _options_levels,
    parse_nabu_config_file,
)
from nabu.stitching.overlap import OverlapStitchingStrategy
from nabu.stitching import config as stiching_config


_stitching_types = list(stiching_config.StitchingType.values())
_stitching_types.append(None)


def nabu_config_to_dict(nabu_config):
    res = {}
    for section, section_content in nabu_config.items():
        res[section] = {}
        for key, values in section_content.items():
            res[section][key] = values["default"]
    return res


@pytest.mark.parametrize("stitching_type", _stitching_types)
@pytest.mark.parametrize("option_level", _options_levels.keys())
def test_stitching_config(stitching_type, option_level):
    """
    insure get_default_stitching_config is returning a dict
    and is coherent with the configuration classes
    """
    with TemporaryDirectory() as output_dir:
        nabu_dict = stiching_config.get_default_stitching_config(stitching_type)
        config = nabu_config_to_dict(nabu_dict)
        assert isinstance(config, dict)

        assert "stitching" in config
        assert "type" in config["stitching"]
        stitching_type = stiching_config.StitchingType.from_value(config["stitching"]["type"])
        if stitching_type is stiching_config.StitchingType.Z_POSTPROC:
            assert isinstance(
                stiching_config.dict_to_config_obj(config),
                stiching_config.PostProcessedSingleAxisStitchingConfiguration,
            )
        elif stitching_type is stiching_config.StitchingType.Z_PREPROC:
            assert isinstance(
                stiching_config.dict_to_config_obj(config),
                stiching_config.PreProcessedSingleAxisStitchingConfiguration,
            )
        elif stitching_type is stiching_config.StitchingType.Y_PREPROC:
            assert isinstance(
                stiching_config.dict_to_config_obj(config),
                stiching_config.PreProcessedSingleAxisStitchingConfiguration,
            )
        else:
            raise ValueError("not handled")

        # dump configuration to file
        output_file = os.path.join(output_dir, "config.conf")
        generate_nabu_configfile(
            fname=output_file,
            default_config=nabu_dict,
            comments=True,
            sections_comments=stiching_config.SECTIONS_COMMENTS,
            options_level=option_level,
            prefilled_values={},
        )

        # load configuration from file
        loaded_config = parse_nabu_config_file(output_file)
        config_class_instance = stiching_config.dict_to_config_obj(loaded_config)
        if stitching_type is stiching_config.StitchingType.Z_POSTPROC:
            assert isinstance(
                config_class_instance,
                stiching_config.PostProcessedSingleAxisStitchingConfiguration,
            )
        elif stitching_type is stiching_config.StitchingType.Z_PREPROC:
            assert isinstance(
                config_class_instance,
                stiching_config.PreProcessedSingleAxisStitchingConfiguration,
            )

        assert isinstance(config_class_instance.to_dict(), dict)


@pytest.mark.parametrize("stitching_strategy", OverlapStitchingStrategy.values())
@pytest.mark.parametrize("overwrite_results", (True, "False", 0, "1"))
@pytest.mark.parametrize(
    "axis_shifts",
    (
        "",
        None,
        "None",
        "",
        "skimage",
        "nabu-fft",
    ),
)
@pytest.mark.parametrize("axis_shifts_params", ("", {}, "window_size=200"))
@pytest.mark.parametrize(
    "slice_for_correlation",
    (
        "middle",
        "3",
    ),
)
@pytest.mark.parametrize("slices", ("middle", "0:26:2"))
@pytest.mark.parametrize(
    "input_scans",
    (
        "",
        "hdf5:scan:/data/scan.hdf5?path=entry; hdf5:scan:/data/scan.hdf5?path=entry1",
    ),
)
@pytest.mark.parametrize(
    "slurm_config",
    (
        {
            stiching_config.SLURM_MODULES_TO_LOADS: "tomotools",
            stiching_config.SLURM_PREPROCESSING_COMMAND: "",
            stiching_config.SLURM_CLEAN_SCRIPTS: True,
            stiching_config.SLURM_MEM: 56,
            stiching_config.SLURM_N_JOBS: 5,
            stiching_config.SLURM_PARTITION: "my_partition",
        },
    ),
)
def test_PreProcessedZStitchingConfiguration(
    stitching_strategy,
    overwrite_results,
    axis_shifts,
    axis_shifts_params,
    input_scans,
    slice_for_correlation,
    slices,
    slurm_config,
):
    """
    make sure configuration works well for PreProcessedZStitchingConfiguration
    """
    pre_process_config = stiching_config.PreProcessedZStitchingConfiguration.from_dict(
        {
            stiching_config.STITCHING_SECTION: {
                stiching_config.CROSS_CORRELATION_SLICE_FIELD: slice_for_correlation,
                stiching_config.AXIS_0_POS_PX: axis_shifts,
                stiching_config.AXIS_1_POS_PX: axis_shifts,
                stiching_config.AXIS_2_POS_PX: axis_shifts,
                stiching_config.AXIS_0_PARAMS: axis_shifts_params,
                stiching_config.AXIS_1_PARAMS: axis_shifts_params,
                stiching_config.AXIS_2_PARAMS: axis_shifts_params,
                stiching_config.STITCHING_STRATEGY_FIELD: stitching_strategy,
            },
            stiching_config.INPUTS_SECTION: {
                stiching_config.INPUT_DATASETS_FIELD: input_scans,
                stiching_config.STITCHING_SLICES: slices,
            },
            stiching_config.OUTPUT_SECTION: {
                stiching_config.OVERWRITE_RESULTS_FIELD: overwrite_results,
            },
            stiching_config.PRE_PROC_SECTION: {
                stiching_config.DATA_FILE_FIELD: "my_file.nx",
                stiching_config.DATA_PATH_FIELD: "entry",
                stiching_config.NEXUS_VERSION_FIELD: None,
            },
            stiching_config.SLURM_SECTION: slurm_config,
            stiching_config.NORMALIZATION_BY_SAMPLE_SECTION: {
                stiching_config.NORMALIZATION_BY_SAMPLE_MARGIN: 1,
                stiching_config.NORMALIZATION_BY_SAMPLE_SIDE: "right",
                stiching_config.NORMALIZATION_BY_SAMPLE_ACTIVE_FIELD: True,
                stiching_config.NORMALIZATION_BY_SAMPLE_METHOD: "mean",
                stiching_config.NORMALIZATION_BY_SAMPLE_WIDTH: 31,
            },
        },
    )

    from_dict = stiching_config.PreProcessedZStitchingConfiguration.from_dict(pre_process_config.to_dict())
    # workaround for scans because a new object is created each time
    pre_process_config.settle_inputs
    assert len(from_dict.input_scans) == len(pre_process_config.input_scans)
    from_dict.input_scans = None
    pre_process_config.input_scans = None
    assert pre_process_config == from_dict


def test_PostProcessedZStitchingConfiguration():
    """
    make sure configuration works well for PostProcessedZStitchingConfiguration
    """
    pass


def test_description_dict():
    """
    make sure the description dict (used for generating the file) is working and generates a dict
    """
    assert isinstance(stiching_config.PreProcessedSingleAxisStitchingConfiguration.get_description_dict(), dict)
    assert isinstance(
        stiching_config.PostProcessedSingleAxisStitchingConfiguration.get_description_dict(),
        dict,
    )



import numpy as np
import pytest
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.opencl.utils import __has_pyopencl__, get_opencl_context
from nabu.processing.padding_cuda import CudaPadding
from nabu.processing.padding_opencl import OpenCLPadding
from nabu.utils import calc_padding_lengths, get_cuda_srcfile
from nabu.testutils import __do_long_tests__
from nabu.testutils import get_data, generate_tests_scenarios

scenarios = {
    "shape": [(511, 512), (512, 511)],
    "pad_width": [((256, 255), (128, 127))],
    "mode_cuda": CudaPadding.supported_modes[:2] if __has_pycuda__ else [],
    "mode_opencl": OpenCLPadding.supported_modes[:2] if __has_pyopencl__ else [],
    "constant_values": [0, ((1.0, 2.0), (3.0, 4.0))],
    "output_is_none": [True, False],
    "backend": ["cuda", "opencl"],
}


if __do_long_tests__:
    scenarios["mode_cuda"] = CudaPadding.supported_modes if __has_pycuda__ else []
    scenarios["mode_opencl"] = OpenCLPadding.supported_modes if __has_pyopencl__ else []
    scenarios["pad_width"].extend([((0, 0), (6, 7))])

scenarios = generate_tests_scenarios(scenarios)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.data = get_data("brain_phantom.npz")["data"]
    cls.tol = 1e-7
    if __has_pycuda__:
        cls.cu_ctx = get_cuda_context(cleanup_at_exit=False)
    if __has_pyopencl__:
        cls.cl_ctx = get_opencl_context(device_type="all")
    yield
    if __has_pycuda__:
        cls.cu_ctx.pop()


@pytest.mark.usefixtures("bootstrap")
class TestPadding:
    @pytest.mark.parametrize("config", scenarios)
    def test_padding(self, config):
        backend = config["backend"]
        shape = config["shape"]
        padding_mode = config["mode_%s" % backend]
        data = self.data[: shape[0], : shape[1]]
        kwargs = {}
        if padding_mode == "constant":
            kwargs["constant_values"] = config["constant_values"]
        ref = np.pad(data, config["pad_width"], mode=padding_mode, **kwargs)

        PaddingCls = CudaPadding if backend == "cuda" else OpenCLPadding
        if backend == "cuda":
            backend_options = {"cuda_options": {"ctx": self.cu_ctx}}
        else:
            backend_options = {"opencl_options": {"ctx": self.cl_ctx}}

        padding = PaddingCls(
            config["shape"],
            config["pad_width"],
            mode=padding_mode,
            constant_values=config["constant_values"],
            **backend_options,
        )
        if config["output_is_none"]:
            output = None
        else:
            output = padding.processing.allocate_array("output", ref.shape, dtype="f")

        d_img = padding.processing.allocate_array("d_img", data.shape, dtype="f")
        d_img.set(np.ascontiguousarray(data, dtype="f"))
        res = padding.pad(d_img, output=output)

        err_max = np.max(np.abs(res.get() - ref))
        assert err_max < self.tol, str("Something wrong with padding for configuration %s" % (str(config)))

    @pytest.mark.skipif(not (__has_pycuda__) and not (__has_pyopencl__), reason="need pycuda or pyopencl")
    def test_custom_coordinate_transform(self):
        data = self.data
        R, C = np.indices(data.shape, dtype=np.int32)

        pad_width = ((256, 255), (254, 251))
        mode = "reflect"

        coords_R = np.pad(R, pad_width[0], mode=mode)[:, 0]
        coords_C = np.pad(C, pad_width[1], mode=mode)[0, :]

        # Further transform of coordinates - here FFT layout
        coords_R = np.roll(coords_R, -pad_width[0][0])
        coords_C = np.roll(coords_C, -pad_width[1][0])

        padding_classes_to_test = []
        if __has_pycuda__:
            padding_classes_to_test.append(CudaPadding)
        if __has_pyopencl__:
            padding_classes_to_test.append(OpenCLPadding)

        for padding_cls in padding_classes_to_test:
            ctx = self.cl_ctx if padding_cls.backend == "opencl" else self.cu_ctx
            padding = padding_cls(data.shape, (coords_R, coords_C), mode=mode, ctx=ctx)

            d_img = padding.processing.allocate_array("d_img", data.shape, dtype="f")
            d_img.set(data)
            d_out = padding.processing.allocate_array("d_out", padding.padded_shape, dtype="f")

            res = padding.pad(d_img, output=d_out)

            ref = np.roll(np.pad(data, pad_width, mode=mode), (-pad_width[0][0], -pad_width[1][0]), axis=(0, 1))

            err_max = np.max(np.abs(d_out.get() - ref))
            assert err_max < self.tol, "Something wrong with custom padding"


#
# The following is testing a previous version of padding kernels
# They use specific code (instead of a generic coordinate transform)
#

if __has_pycuda__:
    from nabu.cuda.kernel import CudaKernel
    import pycuda.gpuarray as garray


scenarios_legacy = [
    {
        "shape": (512, 501),
        "shape_padded": (1023, 1022),
        "constant_values": ((1.0, 2.0), (3.0, 4.0)),
    },
]


# parametrize with fixture and "params=" will launch a new class for each scenario.
# the attributes set to "cls" will remain for all the tests done in this class
# with the current scenario.
@pytest.fixture(scope="class", params=scenarios_legacy)
def bootstrap_legacy(request):
    cls = request.cls
    cls.data = get_data("mri_proj_astra.npz")["data"]
    cls.tol = 1e-7
    cls.params = request.param
    cls.ctx = get_cuda_context(cleanup_at_exit=False)
    cls._calc_pad()
    cls._init_kernels()
    yield
    cls.ctx.pop()


@pytest.mark.skipif(not (__has_pycuda__), reason="Need Cuda and pycuda for this test")
@pytest.mark.usefixtures("bootstrap_legacy")
class TestPaddingLegacy:
    @classmethod
    def _calc_pad(cls):
        cls.shape = cls.params["shape"]
        cls.data = np.ascontiguousarray(cls.data[: cls.shape[0], : cls.shape[1]])
        cls.shape_padded = cls.params["shape_padded"]
        ((pt, pb), (pl, pr)) = calc_padding_lengths(cls.shape, cls.shape_padded)
        cls.pad_top_len = pt
        cls.pad_bottom_len = pb
        cls.pad_left_len = pl
        cls.pad_right_len = pr

    @classmethod
    def _init_kernels(cls):
        cls.pad_kern = CudaKernel(
            "padding_constant",
            filename=get_cuda_srcfile("padding.cu"),
            signature="Piiiiiiiiffff",
        )
        cls.pad_edge_kern = CudaKernel(
            "padding_edge",
            filename=get_cuda_srcfile("padding.cu"),
            signature="Piiiiiiii",
        )
        cls.d_data_padded = garray.zeros(cls.shape_padded, "f")

    def _init_padding(self, arr=None):
        arr = arr or self.data
        self.d_data_padded.fill(0)
        Ny, Nx = self.shape
        self.d_data_padded[:Ny, :Nx] = self.data

    def _pad_numpy(self, arr=None, **np_pad_kwargs):
        arr = arr or self.data
        data_padded_ref = np.pad(
            arr, ((self.pad_top_len, self.pad_bottom_len), (self.pad_left_len, self.pad_right_len)), **np_pad_kwargs
        )
        # Put in the FFT layout
        data_padded_ref = np.roll(data_padded_ref, (-self.pad_top_len, -self.pad_left_len), axis=(0, 1))
        return data_padded_ref

    def test_constant_padding(self):
        self._init_padding()
        # Pad using the cuda kernel
        ((val_top, val_bottom), (val_left, val_right)) = self.params["constant_values"]
        Ny, Nx = self.shape
        Nyp, Nxp = self.shape_padded

        self.pad_kern(
            self.d_data_padded,
            Nx,
            Ny,
            Nxp,
            Nyp,
            self.pad_left_len,
            self.pad_right_len,
            self.pad_top_len,
            self.pad_bottom_len,
            val_left,
            val_right,
            val_top,
            val_bottom,
        )
        # Pad using numpy
        data_padded_ref = self._pad_numpy(mode="constant", constant_values=self.params["constant_values"])
        # Compare
        errmax = np.max(np.abs(self.d_data_padded.get() - data_padded_ref))
        assert errmax < self.tol, "Max error is too high"

    def test_edge_padding(self):
        self._init_padding()
        # Pad using the cuda kernel
        ((val_top, val_bottom), (val_left, val_right)) = self.params["constant_values"]
        Ny, Nx = self.shape
        Nyp, Nxp = self.shape_padded

        self.pad_edge_kern(
            self.d_data_padded,
            Nx,
            Ny,
            Nxp,
            Nyp,
            self.pad_left_len,
            self.pad_right_len,
            self.pad_top_len,
            self.pad_bottom_len,
        )
        # Pad using numpy
        data_padded_ref = self._pad_numpy(mode="edge")
        # Compare
        errmax = np.max(np.abs(self.d_data_padded.get() - data_padded_ref))
        assert errmax < self.tol, "Max error is too high"



import logging
from pprint import pformat

from tqdm import tqdm
from nabu.stitching.slurm_utils import split_stitching_configuration_to_slurm_job
from .cli_configs import StitchingConfig
from ..pipeline.config import parse_nabu_config_file
from nabu.stitching.single_axis_stitching import stitching
from nabu.stitching.utils.post_processing import StitchingPostProcAggregation
from nabu.stitching.config import dict_to_config_obj
from .utils import parse_params_values

try:
    from sluurp.executor import submit
except ImportError:
    has_sluurp = False
else:
    has_sluurp = True

_logger = logging.getLogger(__name__)


def main():
    args = parse_params_values(
        StitchingConfig,
        parser_description="Run stitching from a configuration file. Configuration can be obtain from `stitching-config`",
    )
    logging.basicConfig(level=args["loglevel"].upper())

    conf_dict = parse_nabu_config_file(args["input-file"], allow_no_value=True)

    stitching_config = dict_to_config_obj(conf_dict)
    assert stitching_config.axis is not None, "axis must be defined to know how to stitch"
    _logger.info(" when loaded axis is %s", stitching_config.axis)
    stitching_config.settle_inputs()
    if args["only_create_master_file"]:
        # option to ease creation of the master in the following cases:
        # * user has submitted all the job but has been kicked out of the cluster
        # * only a few slurm job for some random version (cluster update...) and user want to retrigger only those job and process the aggregation only. On those cases no need to redo it all.
        tomo_objs = []
        for _, sub_config in split_stitching_configuration_to_slurm_job(stitching_config, yield_configuration=True):
            tomo_objs.append(sub_config.get_output_object().get_identifier().to_str())

        post_processing = StitchingPostProcAggregation(
            existing_objs_ids=tomo_objs,
            stitching_config=stitching_config,
        )
        post_processing.process()

    elif stitching_config.slurm_config.partition in (None, ""):
        # case 1: run locally
        _logger.info("run stitching locally with: %s", pformat(stitching_config.to_dict()))

        main_progress = tqdm(total=100, desc="stitching", leave=True)
        stitching(stitching_config, progress=main_progress)
    else:
        if not has_sluurp:
            raise ImportError(
                "sluurp not install. Please install it to distribute stitching on slurm (pip install slurm)"
            )
        main_progress = tqdm(total=100, position=0, desc="stitching")

        # case 2: run on slurm
        # note: to speed up we could do shift research on pre processing and run it only once (if manual of course). Here it will be run for all part
        _logger.info(f"will distribute stitching")

        futures = {}
        # 2.1 launch jobs
        slurm_job_progress_bars: dict = {}
        for i_job, (job, sub_config) in enumerate(
            split_stitching_configuration_to_slurm_job(stitching_config, yield_configuration=True)
        ):
            _logger.info(f"submit job nb {i_job}: handles {sub_config.slices}")
            output_volume = sub_config.get_output_object().get_identifier().to_str()
            futures[output_volume] = submit(job, timeout=999999)
            # note on total=100: we only consider percentage in this case (providing advancement from slurm jobs)
            slurm_job_progress_bars[job] = tqdm(
                total=100,
                position=i_job + 1,
                desc=f"   part {str(i_job).ljust(3)}",
                delay=0.5,  # avoid to mess with terminal and (near) future logs
                bar_format="{l_bar}{bar}",  # avoid using 'r_bar' as 'total' is set to 100 (percentage)
                leave=False,
            )

        main_progress.n = 50
        # 2.2 wait for future to be done and concatenate the result
        post_processing = StitchingPostProcAggregation(
            futures=futures,
            stitching_config=stitching_config,
            progress_bars=slurm_job_progress_bars,
        )
        post_processing.process()

    exit(0)


if __name__ == "__main__":
    main()



import numpy as np
from ..utils import updiv, nextpow2, convert_index, deprecation_warning
from ..processing.processing_base import ProcessingBase
from .filtering import SinoFilter
from .sinogram import SinoMult
from .sinogram import get_extended_sinogram_width


class BackprojectorBase:
    """
    Base class for backprojectors.
    """

    backend = "numpy"
    default_padding_mode = "zeros"
    kernel_name = "backproj"
    default_extra_options = {
        "padding_mode": None,
        "axis_correction": None,
        "centered_axis": False,
        "clip_outer_circle": False,
        "scale_factor": None,
        "filter_cutoff": 1.0,
        "outer_circle_value": 0.0,
    }

    kernel_filename = None
    backend_processing_class = ProcessingBase
    SinoFilterClass = SinoFilter
    SinoMultClass = SinoMult
    _sino_filter_other_options = {}

    def __init__(
        self,
        sino_shape,
        slice_shape=None,
        angles=None,
        rot_center=None,
        padding_mode=None,
        halftomo=False,
        filter_name=None,
        slice_roi=None,
        scale_factor=None,
        extra_options=None,
        backend_options=None,
    ):
        """
        Initialize a Backprojector.

        Parameters
        -----------
        sino_shape: tuple
            Shape of the sinogram, in the form `(n_angles, detector_width)`
            (for backprojecting one sinogram) or `(n_sinos, n_angles, detector_width)`.
        slice_shape: int or tuple, optional
            Shape of the slice. By default, the slice shape is (n_x, n_x) where
            `n_x = detector_width`
        angles: array-like, optional
            Rotation anles in radians.
            By default, angles are equispaced between [0, pi[.
        rot_center: float, optional
            Rotation axis position. Default is `(detector_width - 1)/2.0`
        padding_mode: str, optional
            Padding mode when filtering the sinogram. Can be "zeros" (default) or "edges".
        filter_name: str, optional
            Name of the filter for filtered-backprojection.
        slice_roi: tuple, optional.
            Whether to backproject in a restricted area.
            If set, it must be in the form (start_x, end_x, start_y, end_y).
            `end_x` and `end_y` are non inclusive ! For example if the detector has
            2048 pixels horizontally, then you can choose `start_x=0` and `end_x=2048`.
            If one of the value is set to None, it is replaced with a default value
            (0 for start, n_x and n_y for end)
        scale_factor: float, optional
            Scaling factor for backprojection.
            For example, to get the linear absorption coefficient in 1/cm,
            this factor has to be set as the pixel size in cm.
            DEPRECATED - please use this parameter in "extra_options"
        extra_options: dict, optional
            Advanced extra options.
             See the "Extra options" section for more information.
        backend_options: dict, optional
            OpenCL/Cuda options passed to the OpenCLProcessing or CudaProcessing class.

        Other parameters
        -----------------
        extra_options: dict, optional
            Dictionary with a set of advanced options. The default are the following:
                - "padding_mode": "zeros"
                   Padding mode when filtering the sinogram. Can be "zeros" or "edges".
                   DEPRECATED - please use "padding_mode" directly in parameters.
                - "axis_correction": None
                    Whether to set a correction for the rotation axis.
                    If set, this should be an array with as many elements as the number
                    of angles. This is useful when there is an horizontal displacement
                    of the rotation axis.
                - centered_axis: bool
                    Whether to "center" the slice on the rotation axis position.
                    If set to True, then the reconstructed region is centered on the rotation axis.
                - scale_factor: float
                    Scaling factor for backprojection.
                    For example, to get the linear absorption coefficient in 1/cm,
                    this factor has to be set as the pixel size in cm.
                - clip_outer_circle: False
                    Whether to set to zero the pixels outside the reconstruction mask
                - filter_cutoff: float
                    Cut-off frequency usef for Fourier filter. Default is 1.0
        """
        self._processing = self.backend_processing_class(**(backend_options or {}))
        self._configure_extra_options(scale_factor, padding_mode, extra_options=extra_options)
        self._check_textures_availability()
        self._init_geometry(sino_shape, slice_shape, angles, rot_center, halftomo, slice_roi)
        self._init_filter(filter_name)
        self._allocate_memory()
        self._compute_angles()
        self._compile_kernels()

    def _configure_extra_options(self, scale_factor, padding_mode, extra_options=None):
        extra_options = extra_options or {}
        # compat.
        scale_factor = None
        if scale_factor is not None:
            deprecation_warning(
                "Please use the parameter 'scale_factor' in the 'extra_options' dict",
                do_print=True,
                func_name="fbp_scale_factor",
            )
        scale_factor = extra_options.get("scale_factor", None) or scale_factor or 1.0
        #
        if "padding_mode" in extra_options:
            deprecation_warning(
                "Please use 'padding_mode' directly in Backprojector arguments, not in 'extra_options'",
                do_print=True,
                func_name="fbp_padding_mode",
            )
        #
        self._backproj_scale_factor = scale_factor
        self._axis_array = None
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options)
        self.padding_mode = padding_mode or self.extra_options["padding_mode"] or self.default_padding_mode
        self._axis_array = self.extra_options["axis_correction"]

    def _init_geometry(self, sino_shape, slice_shape, angles, rot_center, halftomo, slice_roi):
        if slice_shape is not None and slice_roi is not None:
            raise ValueError("slice_shape and slice_roi cannot be used together")
        self.sino_shape = sino_shape
        if len(sino_shape) == 2:
            n_angles, dwidth = sino_shape
        else:
            raise ValueError("Expected 2D sinogram")
        self.dwidth = dwidth
        self.halftomo = halftomo
        if rot_center is None:
            if halftomo:
                raise ValueError("Need to know 'rot_center' when using halftomo")
            rot_center = (self.dwidth - 1) / 2.0
        self.rot_center = rot_center
        self._set_slice_shape(slice_shape)
        self.axis_pos = self.rot_center
        self._set_angles(angles, n_angles)
        self._set_slice_roi(slice_roi)
        #
        # offset = start - move
        # move = 0 if not(centered_axis) else start + (n-1)/2. - c
        if self.extra_options["centered_axis"]:
            self.offsets = {
                "x": self.rot_center - (self.n_x - 1) / 2.0,
                "y": self.rot_center - (self.n_y - 1) / 2.0,
            }
        #
        self._set_axis_corr()

    def _set_slice_shape(self, slice_shape):
        if not (self.halftomo):
            n_x = n_y = self.dwidth
        else:
            n_x = n_y = get_extended_sinogram_width(self.dwidth, self.rot_center)
        if slice_shape is not None:
            if np.isscalar(slice_shape):
                slice_shape = (slice_shape, slice_shape)
            n_y, n_x = slice_shape
        self.n_x = n_x
        self.n_y = n_y
        self.slice_shape = (n_y, n_x)

    def _set_angles(self, angles, n_angles):
        self.n_angles = n_angles
        if angles is None:
            angles = n_angles
        if np.isscalar(angles):
            end_angle = np.pi if not (self.halftomo) else 2 * np.pi
            take_end_angle = self.halftomo
            angles = np.linspace(0, end_angle, angles, take_end_angle)
        else:
            assert len(angles) == self.n_angles, "expected %d angles but got %d" % (len(angles), self.n_angles)
        self.angles = angles

    def _set_slice_roi(self, slice_roi):
        self.offsets = {"x": 0, "y": 0}
        self.slice_roi = slice_roi
        if slice_roi is None:
            return
        start_x, end_x, start_y, end_y = slice_roi
        # convert negative indices
        start_x = convert_index(start_x, self.n_x, 0)
        start_y = convert_index(start_y, self.n_y, 0)
        end_x = convert_index(end_x, self.n_x, self.n_x)
        end_y = convert_index(end_y, self.n_y, self.n_y)
        self.slice_shape = (end_y - start_y, end_x - start_x)
        self.n_x = self.slice_shape[-1]
        self.n_y = self.slice_shape[-2]
        self.offsets = {"x": start_x, "y": start_y}

    def _allocate_memory(self):
        # 1D textures are not supported in pyopencl
        self.h_msin = np.zeros((1, self.n_angles), "f")
        self.h_cos = np.zeros((1, self.n_angles), "f")
        # self._d_sino = self._processing.allocate_array("d_sino", self.sino_shape, "f")
        self._processing.init_arrays_to_none(["_d_slice", "d_sino"])

    def _compute_angles(self):
        self.h_cos[0] = np.cos(self.angles).astype("f")
        self.h_msin[0] = (-np.sin(self.angles)).astype("f")
        self._d_msin = self._processing.set_array("d_msin", self.h_msin[0])
        self._d_cos = self._processing.set_array("d_cos", self.h_cos[0])
        if self._axis_correction is not None:
            self._d_axcorr = self._processing._set_array("d_axcorr", self._axis_correction)

    def _set_axis_corr(self):
        axcorr = self.extra_options["axis_correction"]
        self._axis_correction = axcorr
        if axcorr is None:
            return
        if len(axcorr) != self.n_angles:
            raise ValueError("Expected %d angles but got %d" % (self.n_angles, len(axcorr)))
        self._axis_correction = np.zeros((1, self.n_angles), dtype=np.float32)
        self._axis_correction[0, :] = axcorr[:]  # pylint: disable=E1136

    def _init_filter(self, filter_name):
        self.filter_name = filter_name
        if filter_name in ["None", "none"]:
            self.sino_filter = None
            return
        sinofilter_other_kwargs = {}
        if self.backend != "numpy":
            sinofilter_other_kwargs["%s_options" % self.backend] = {"ctx": self._processing.ctx}
        self.sino_filter = self.SinoFilterClass(
            self.sino_shape,
            filter_name=self.filter_name,
            padding_mode=self.padding_mode,
            extra_options={"cutoff": self.extra_options.get("filter_cutoff", 1.0)},
            **sinofilter_other_kwargs,
        )
        if self.halftomo:
            # When doing half-tomography, each projections is seen "twice".
            # SinoFilter normalizes with pi/n_angles, but in half-tomography here n_angles is somehow halved.
            # TODO it should even be "n_turns", where n_turns can be computed from the angles
            self.sino_filter.set_filter(self.sino_filter.filter_f * (self.n_angles / np.pi * 2))

    def reset_rot_center(self, rot_center):
        """
        Define a new center of rotation for the current backprojector.
        """
        self.rot_center = rot_center
        self.axis_pos = rot_center
        # See kernels signature of backproj.cu and backproj.cl.
        # The ifdef makes things a bit more complicated
        proj_arg_idx = 4
        if self.backend == "cuda" and self._use_textures:
            proj_arg_idx = 3
        self.kern_proj_args[proj_arg_idx] = rot_center
        if self.extra_options["centered_axis"]:
            self.offsets = {
                "x": self.rot_center - (self.n_x - 1) / 2.0,
                "y": self.rot_center - (self.n_y - 1) / 2.0,
            }
            self.kern_proj_args[proj_arg_idx + 3] = self.offsets["x"]
            self.kern_proj_args[proj_arg_idx + 4] = self.offsets["y"]

    # Try to factorize some code between Cuda and OpenCL
    # Not ideal, as cuda uses "grid" = n_blocks_launched,
    # while OpenCL uses "global_size" = n_threads_launched
    def _get_kernel_options(self):
        sourcemodule_options = []
        # We use blocks of 16*16 (see why in kernel doc), and one thread
        # handles 2 pixels per dimension.
        block = (16, 16, 1)
        # The Cuda kernel is optimized for 16x16 threads blocks
        # If one of the dimensions is smaller than 16, it has to be addapted
        if self.n_x < 16 or self.n_y < 16:
            tpb_x = min(int(nextpow2(self.n_x)), 16)
            tpb_y = min(int(nextpow2(self.n_y)), 16)
            block = (tpb_x, tpb_y, 1)
            sourcemodule_options.append("-DSHARED_SIZE=%d" % (tpb_x * tpb_y))
        grid = (updiv(updiv(self.n_x, block[0]), 2), updiv(updiv(self.n_y, block[1]), 2))
        if self.extra_options["clip_outer_circle"]:
            sourcemodule_options.append("-DCLIP_OUTER_CIRCLE")
        shared_size = int(np.prod(block)) * 2
        if self._axis_correction is not None:
            sourcemodule_options.append("-DDO_AXIS_CORRECTION")
            shared_size += int(np.prod(block))
        shared_size *= 4  # sizeof(float32)
        self._kernel_options = {
            "kernel_name": self.kernel_name,
            "sourcemodule_options": sourcemodule_options,
            "grid": grid,
            "block": block,
            "shared_size": shared_size,
        }

    def _prepare_kernel_args(self):
        self._get_kernel_options()
        self.kern_proj_args = [
            None,  # output d_slice holder
            None,  # placeholder for sino (OpenCL or Cuda+no-texture)
            np.int32(self.n_angles),
            np.int32(self.dwidth),
            np.float32(self.axis_pos),
            np.int32(self.n_x),
            np.int32(self.n_y),
            np.float32(self.offsets["x"]),
            np.float32(self.offsets["y"]),
            self._d_cos,
            self._d_msin,
            np.float32(self._backproj_scale_factor),
        ]
        if self._axis_correction is not None:
            self.kern_proj_args.insert(-1, self._d_axcorr)
        self.kern_proj_kwargs = {
            "grid": self._kernel_options["grid"],
            "block": self._kernel_options["block"],
        }

    def _set_output(self, output, check=False):
        self._output_is_ndarray = isinstance(output, np.ndarray)
        if output is None or self._output_is_ndarray:
            self._processing.allocate_array("_d_slice", self.slice_shape, dtype=np.float32)
            output = self._processing._d_slice  # pylint: disable=E1101
        elif check:
            assert output.dtype == np.float32
            assert output.shape == self.slice_shape, "Expected output shape %s but got %s" % (
                self.slice_shape,
                output.shape,
            )
        if self.extra_options.get("clip_outer_circle", False):
            out_circle_val = self.extra_options.get("outer_circle_value", 0)
            if out_circle_val != 0:
                output.fill(out_circle_val)
        return output

    def _set_kernel_slice_arg(self, d_slice):
        self.kern_proj_args[0] = d_slice

    def backproj(self, sino, output=None, do_checks=True):
        if self.halftomo and self.rot_center < self.dwidth:
            self.sino_mult.prepare_sino(sino)
        self._transfer_to_texture(sino)
        d_slice = self._set_output(output, check=do_checks)
        self._set_kernel_slice_arg(d_slice)
        self.gpu_projector(*self.kern_proj_args, **self.kern_proj_kwargs)
        if output is not None and not (self._output_is_ndarray):
            return output
        else:
            return self._processing._d_slice.get(ary=output)

    def filtered_backprojection(self, sino, output=None):
        #
        if isinstance(sino, self._processing.array_class):
            d_sino = sino
        else:
            d_sino = self._processing.to_device("d_sino", sino)
        #
        if self.sino_filter is not None:
            filt_kwargs = {}
            # if a new device array was allocated for sinogram, then the filtering can overwrite it,
            # since it won't affect user argument
            if id(d_sino) != id(sino):
                filt_kwargs = {"output": d_sino}
            #
            sino_to_backproject = self.sino_filter(d_sino, **filt_kwargs)
        else:
            sino_to_backproject = d_sino
        return self.backproj(sino_to_backproject, output=output)

    fbp = filtered_backprojection  # shorthand

    def __repr__(self):
        res = "%s(sino_shape=%s, slice_shape=%s, rot_center=%.2f, halftomo=%s)" % (
            self.__class__.__name__,
            str(self.sino_shape),
            str(self.slice_shape),
            self.rot_center,
            self.halftomo,
        )
        return res



from ..processing.rotation_cuda import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.rotation_cuda has been moved to nabu.processing.rotation_cuda", do_print=True, func_name="rotation_cuda"
)



import numpy as np
from ..opencl.kernel import OpenCLKernel
from ..opencl.processing import OpenCLProcessing
from ..utils import get_opencl_srcfile
from .sinogram import SinoMult


class OpenCLSinoMult(SinoMult):
    def __init__(self, sino_shape, rot_center, **opencl_options):
        super().__init__(sino_shape, rot_center)
        self.opencl_processing = OpenCLProcessing(**opencl_options)
        self._init_kernel()

    def _init_kernel(self):
        self.halftomo_kernel = OpenCLKernel(
            "halftomo_prepare_sinogram",
            self.opencl_processing.ctx,
            filename=get_opencl_srcfile("halftomo.cl"),
        )
        self.d_weights = self.opencl_processing.set_array("d_weights", self.weights)
        self._halftomo_kernel_other_args = [
            self.d_weights,
            np.int32(self.n_a),
            np.int32(self.n_x),
            np.int32(self.start_x),
            np.int32(self.end_x),
        ]
        self._global_size = (self.n_x, self.n_a)
        self._local_size = None  # (32, 32, 1)  # tune ?

    def prepare_sino(self, sino):
        sino = self.opencl_processing.set_array("d_sino", sino)
        ev = self.halftomo_kernel(
            self.opencl_processing.queue,
            sino,
            *self._halftomo_kernel_other_args,
            global_size=self._global_size,
            local_size=self._local_size,
        )
        if self.opencl_processing.device_type == "cpu":
            ev.wait()
        return sino



import numpy as np
from scipy.interpolate import RegularGridInterpolator
from ..utils import check_supported
from ..estimation.distortion import estimate_flat_distortion


def correct_distortion_interpn(image, coords, bounds_error=False, fill_value=None):
    """
    Correct image distortion with scipy.interpolate.interpn.

    Parameters
    ----------
    image: array
        Distorted image
    coords: array
        Coordinates of the distortion correction to apply, with the shape (Ny, Nx, 2)
    """
    foo = RegularGridInterpolator(
        (np.arange(image.shape[0]), np.arange(image.shape[1])),
        image,
        bounds_error=bounds_error,
        method="linear",
        fill_value=fill_value,
    )
    return foo(coords)


class DistortionCorrection:
    """
    A class for estimating and correcting image distortion.
    """

    estimation_methods = {
        "fft-correlation": estimate_flat_distortion,
    }
    correction_methods = {
        "interpn": correct_distortion_interpn,
    }

    def __init__(
        self,
        estimation_method="fft-correlation",
        estimation_kwargs=None,
        correction_method="interpn",
        correction_kwargs=None,
    ):
        """
        Initialize a DistortionCorrection object.

        Parameters
        -----------
        estimation_method: str
            Name of the method to use for estimating the distortion
        estimation_kwargs: dict, optional
            Named arguments to pass to the estimation method, in the form of a dictionary.
        correction_method: str
            Name of the method to use for correcting the distortion
        correction_kwargs: dict, optional
            Named arguments to pass to the correction method, in the form of a dictionary.
        """
        self._set_estimator(estimation_method, estimation_kwargs)
        self._set_corrector(correction_method, correction_kwargs)

    def _set_estimator(self, estimation_method, estimation_kwargs):
        check_supported(estimation_method, self.estimation_methods.keys(), "estimation method")
        self.estimator = self.estimation_methods[estimation_method]
        self._estimator_kwargs = estimation_kwargs or {}

    def _set_corrector(self, correction_method, correction_kwargs):
        check_supported(correction_method, self.correction_methods.keys(), "correction method")
        self.corrector = self.correction_methods[correction_method]
        self._corrector_kwargs = correction_kwargs or {}

    def estimate_distortion(self, image, reference_image):
        return self.estimator(image, reference_image, **self._estimator_kwargs)

    estimate = estimate_distortion

    def correct_distortion(self, image, coords):
        image_corrected = self.corrector(image, coords, **self._corrector_kwargs)
        fill_value = self._corrector_kwargs.get("fill_value", None)
        if fill_value is not None and np.isnan(fill_value):
            mask = np.isnan(image_corrected)
            image_corrected[mask] = image[mask]
        return image_corrected

    correct = correct_distortion

    def estimate_and_correct(self, image, reference_image):
        coords = self.estimate_distortion(image, reference_image)
        image_corrected = self.correct_distortion(image, coords)
        return image_corrected



from .ccd import CCDFilter, Log
from .ctf import CTFPhaseRetrieval
from .distortion import DistortionCorrection
from .double_flatfield import DoubleFlatField
from .flatfield import FlatField, FlatFieldDataUrls
from .phase import PaganinPhaseRetrieval
from .shift import VerticalShift



from multiprocessing.pool import ThreadPool
from bisect import bisect_left
import numpy as np
from ..io.reader import load_images_from_dataurl_dict
from ..utils import check_supported, deprecated_class, get_num_threads


class FlatFieldArrays:
    """
    A class for flat-field normalization
    """

    # the variable below will be True for the derived class
    # which is taylored for to helical case
    _full_shape = False

    _supported_interpolations = ["linear", "nearest"]

    def __init__(
        self,
        radios_shape: tuple,
        flats,
        darks,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
        nan_value=1.0,
        radios_srcurrent=None,
        flats_srcurrent=None,
        n_threads=None,
    ):
        """
        Initialize a flat-field normalization process.

        Parameters
        ----------
        radios_shape: tuple
            A tuple describing the shape of the radios stack, in the form
            `(n_radios, n_z, n_x)`.
        flats: dict
            Dictionary where each key is the flat index, and the value is a
            numpy.ndarray of the flat image.
        darks: dict
            Dictionary where each key is the dark index, and the value is a
            numpy.ndarray of the dark image.
        radios_indices: array of int, optional
            Array containing the radios indices in the scan. `radios_indices[0]` is the index
            of the first radio, and so on.
        interpolation: str, optional
            Interpolation method for flat-field. See below for more details.
        distortion_correction: DistortionCorrection, optional
            A DistortionCorrection object. If provided, it is used to correct flat distortions based on each radio.
        nan_value: float, optional
            Which float value is used to replace nan/inf after flat-field.
        radios_srcurrent: array, optional
            Array with the same shape as radios_indices. Each item contains the synchrotron electric current.
            If not None, normalization with current is applied.
            Please refer to "Notes" for more information on this normalization.
        flats_srcurrent: array, optional
            Array with the same length as "flats". Each item is a measurement of the synchrotron electric current
            for the corresponding flat. The items must be ordered in the same order as the flats indices (`flats.keys()`).
            This parameter must be used along with 'radios_srcurrent'.
            Please refer to "Notes" for more information on this normalization.
        n_threads: int or None, optional
            Number of threads to use for flat-field correction. Default is to use half the threads.

        Important
        ----------
        `flats` and `darks` are expected to be a dictionary with integer keys (the flats/darks indices)
        and numpy array values.
        You can use the following helper functions: `nabu.io.reader.load_images_from_dataurl_dict`
        and `nabu.io.utils.create_dict_of_indices`


        Notes
        ------
        Usually, when doing a scan, only one or a few darks/flats are acquired.
        However, the flat-field normalization has to be performed on each radio,
        although incoming beam can fluctuate between projections.
        The usual way to overcome this is to interpolate between flats.
        If interpolation="nearest", the first flat is used for the first
        radios subset, the second flat is used for the second radios subset,
        and so on.
        If interpolation="linear", the normalization is done as a linear
        function of the radio index.

        The normalization with synchrotron electric current is done as follows.
        Let s = sr/sr_max denote the ratio between current and maximum current,
        D be the dark-current frame, and X' be the normalized frame. Then:
          srcurrent_normalization(X) = X' = (X - D)/s + D
          flatfield_normalization(X') = (X' - D)/(F' - D) = (X - D) / (F - D) * sF/sX
        So current normalization boils down to a scalar multiplication after flat-field.
        """
        if self._full_shape:
            # this is never going to happen in this base class. But in the derived class for helical
            # which needs to keep the full shape
            if radios_indices is not None:
                radios_shape = (len(radios_indices),) + radios_shape[1:]

        self._set_parameters(radios_shape, radios_indices, interpolation, nan_value)
        self._set_flats_and_darks(flats, darks)
        self._precompute_flats_indices_weights()
        self._configure_srcurrent_normalization(radios_srcurrent, flats_srcurrent)
        self.distortion_correction = distortion_correction
        self.n_threads = min(1, get_num_threads(n_threads) // 2)

    def _set_parameters(self, radios_shape, radios_indices, interpolation, nan_value):
        self._set_radios_shape(radios_shape)
        if radios_indices is None:
            radios_indices = np.arange(0, self.n_radios, dtype=np.int32)
        else:
            radios_indices = np.array(radios_indices, dtype=np.int32)
            self._check_radios_and_indices_congruence(radios_indices)

        self.radios_indices = radios_indices
        self.interpolation = interpolation
        check_supported(interpolation, self._supported_interpolations, "Interpolation mode")
        self.nan_value = nan_value
        self._radios_idx_to_pos = dict(zip(self.radios_indices, np.arange(self.radios_indices.size)))

    def _set_radios_shape(self, radios_shape):
        if len(radios_shape) == 2:
            self.radios_shape = (1,) + radios_shape
        elif len(radios_shape) == 3:
            self.radios_shape = radios_shape
        else:
            raise ValueError("Expected radios to have 2 or 3 dimensions")
        n_radios, n_z, n_x = self.radios_shape
        self.n_radios = n_radios
        self.n_angles = n_radios
        self.shape = (n_z, n_x)

    def _set_flats_and_darks(self, flats, darks):
        self._check_frames(flats, "flats", 1, 9999)
        self.n_flats = len(flats)
        self.flats = flats
        self._sorted_flat_indices = sorted(self.flats.keys())

        if self._full_shape:
            # this is never going to happen in this base class. But in the derived class for helical
            # which needs to keep the full shape
            self.shape = flats[self._sorted_flat_indices[0]].shape

        self._flat2arrayidx = dict(zip(self._sorted_flat_indices, np.arange(self.n_flats)))
        self.flats_arr = np.zeros((self.n_flats,) + self.shape, "f")
        for i, idx in enumerate(self._sorted_flat_indices):
            self.flats_arr[i] = self.flats[idx]

        self._check_frames(darks, "darks", 1, 1)
        self.darks = darks
        self.n_darks = len(darks)
        self._sorted_dark_indices = sorted(self.darks.keys())
        self._dark = None

    def _check_frames(self, frames, frames_type, min_frames_required, max_frames_supported):
        n_frames = len(frames)
        if n_frames < min_frames_required:
            raise ValueError("Need at least %d %s" % (min_frames_required, frames_type))
        if n_frames > max_frames_supported:
            raise ValueError(
                "Flat-fielding with more than %d %s is not supported" % (max_frames_supported, frames_type)
            )
        self._check_frame_shape(frames, frames_type)

    def _check_frame_shape(self, frames, frames_type):
        for frame_idx, frame in frames.items():
            if frame.shape != self.shape:
                raise ValueError(
                    "Invalid shape for %s %s: expected %s, but got %s"
                    % (frames_type, frame_idx, str(self.shape), str(frame.shape))
                )

    def _check_radios_and_indices_congruence(self, radios_indices):
        if radios_indices.size != self.n_radios:
            raise ValueError(
                "Expected radios_indices to have length %s = n_radios, but got length %d"
                % (self.n_radios, radios_indices.size)
            )

    def _precompute_flats_indices_weights(self):
        """
        Build two arrays: "indices" and "weights".
        These arrays contain pre-computed information so that the interpolated flat is obtained with

           flat_interpolated = weight_prev * flat_prev  + weight_next * flat_next

        where
           weight_prev, weight_next = weights[2*i], weights[2*i+1]
           idx_prev, idx_next = indices[2*i], indices[2*i+1]
           flat_prev, flat_next = flats[idx_prev], flats[idx_next]

        In words:
          - If a projection has an index between two flats, the equivalent flat is a linear interpolation
            between "previous flat" and "next flat".
          - If a projection has the same index as a flat, only this flat is used for normalization
            (this case normally never occurs, but it's handled in the code)
        """

        def _interp_linear(idx, prev_next):
            if len(prev_next) == 1:  # current index corresponds to an acquired flat
                weights = (1, 0)
                f_idx = (self._flat2arrayidx[prev_next[0]], -1)
            else:
                prev_idx, next_idx = prev_next
                delta = next_idx - prev_idx
                w1 = 1 - (idx - prev_idx) / delta
                w2 = 1 - (next_idx - idx) / delta
                weights = (w1, w2)
                f_idx = (self._flat2arrayidx[prev_idx], self._flat2arrayidx[next_idx])
            return f_idx, weights

        def _interp_nearest(idx, prev_next):
            if len(prev_next) == 1:  # current index corresponds to an acquired flat
                weights = (1, 0)
                f_idx = (self._flat2arrayidx[prev_next[0]], -1)
            else:
                prev_idx, next_idx = prev_next
                idx_to_take = prev_idx if abs(idx - prev_idx) < abs(idx - next_idx) else next_idx
                weights = (1, 0)
                f_idx = (self._flat2arrayidx[idx_to_take], -1)
            return f_idx, weights

        self.flats_idx = np.zeros((self.n_radios, 2), dtype=np.int32)
        self.flats_weights = np.zeros((self.n_radios, 2), dtype=np.float32)
        for i, idx in enumerate(self.radios_indices):
            prev_next = self.get_previous_next_indices(self._sorted_flat_indices, idx)
            if self.interpolation == "nearest":
                f_idx, weights = _interp_nearest(idx, prev_next)
            elif self.interpolation == "linear":
                f_idx, weights = _interp_linear(idx, prev_next)
            # pylint: disable=E0606
            self.flats_idx[i] = f_idx
            self.flats_weights[i] = weights

    # pylint: disable=E1307
    def _configure_srcurrent_normalization(self, radios_srcurrent, flats_srcurrent):
        self.normalize_srcurrent = False
        if radios_srcurrent is None or flats_srcurrent is None:
            return
        radios_srcurrent = np.array(radios_srcurrent)
        if radios_srcurrent.size != self.n_radios:
            raise ValueError(
                "Expected 'radios_srcurrent' to have %d elements but got %d" % (self.n_radios, radios_srcurrent.size)
            )
        flats_srcurrent = np.array(flats_srcurrent)
        if flats_srcurrent.size != self.n_flats:
            raise ValueError(
                "Expected 'flats_srcurrent' to have %d elements but got %d" % (self.n_flats, flats_srcurrent.size)
            )
        self.normalize_srcurrent = True
        self.radios_srcurrent = radios_srcurrent
        self.flats_srcurrent = flats_srcurrent
        self.srcurrent_ratios = np.zeros(self.n_radios, "f")
        # Flats SRCurrent is obtained with "nearest" interp, to emulate an already-done flats SR current normalization
        for i, radio_idx in enumerate(self.radios_indices):
            flat_idx = self.get_nearest_index(self._sorted_flat_indices, radio_idx)
            flat_srcurrent = self.flats_srcurrent[self._flat2arrayidx[flat_idx]]
            self.srcurrent_ratios[i] = flat_srcurrent / self.radios_srcurrent[i]

    @staticmethod
    def get_previous_next_indices(arr, idx):
        pos = bisect_left(arr, idx)
        if pos == len(arr):  # outside range
            return (arr[-1],)
        if arr[pos] == idx:
            return (idx,)
        if pos == 0:
            return (arr[0],)
        return arr[pos - 1], arr[pos]

    @staticmethod
    def get_nearest_index(arr, idx):
        pos = bisect_left(arr, idx)
        if pos == len(arr) or arr[pos] == idx:
            return arr[-1]
        return arr[pos - 1] if idx - arr[pos - 1] < arr[pos] - idx else arr[pos]

    @staticmethod
    def interp(pos, indices, weights, array, slice_y=slice(None, None), slice_x=slice(None, None)):
        """
        Interpolate between two values. The interpolator consists in pre-computed arrays such that

           prev, next = indices[pos]
           w1, w2 = weights[pos]
           interpolated_value = w1 * array[prev] + w2 * array[next]
        """
        prev_idx = indices[pos, 0]
        next_idx = indices[pos, 1]

        if slice_y != slice(None, None) or slice_x != slice(None, None):
            w1 = weights[pos, 0][slice_y, slice_x]
            w2 = weights[pos, 1][slice_y, slice_x]
        else:
            w1 = weights[pos, 0]
            w2 = weights[pos, 1]

        if next_idx == -1:
            val = array[prev_idx]
        else:
            val = w1 * array[prev_idx] + w2 * array[next_idx]
        return val

    def get_flat(self, pos, dtype=np.float32, slice_y=slice(None, None), slice_x=slice(None, None)):
        flat = self.interp(pos, self.flats_idx, self.flats_weights, self.flats_arr, slice_y=slice_y, slice_x=slice_x)
        if flat.dtype != dtype:
            flat = np.ascontiguousarray(flat, dtype=dtype)
        return flat

    def get_dark(self):
        if self._dark is None:
            first_dark_idx = self._sorted_dark_indices[0]
            dark = np.ascontiguousarray(self.darks[first_dark_idx], dtype=np.float32)
            self._dark = dark
        return self._dark

    def remove_invalid_values(self, img):
        if self.nan_value is None:
            return
        invalid_mask = np.logical_not(np.isfinite(img))
        img[invalid_mask] = self.nan_value

    def normalize_radios(self, radios):
        """
        Apply a flat-field normalization, with the current parameters, to a stack
        of radios.
        The processing is done in-place, meaning that the radios content is overwritten.

        Parameters
        -----------
        radios: numpy.ndarray
            Radios chunk
        """
        do_flats_distortion_correction = self.distortion_correction is not None
        dark = self.get_dark()

        def apply_flatfield(i):
            radio_data = radios[i]
            radio_data -= dark
            flat = self.get_flat(i)
            flat = flat - dark
            if do_flats_distortion_correction:
                flat = self.distortion_correction.estimate_and_correct(flat, radio_data)
            np.divide(radio_data, flat, out=radio_data)
            self.remove_invalid_values(radio_data)

        if self.n_threads > 2:
            with ThreadPool(self.n_threads) as tp:
                tp.map(apply_flatfield, range(self.n_radios))
        else:
            for i in range(self.n_radios):
                apply_flatfield(i)

        if self.normalize_srcurrent:
            radios *= self.srcurrent_ratios[:, np.newaxis, np.newaxis]
        return radios

    def normalize_single_radio(
        self, radio, radio_idx, dtype=np.float32, slice_y=slice(None, None), slice_x=slice(None, None)
    ):
        """
        Apply a flat-field normalization to a single projection image.
        """
        dark = self.get_dark()[slice_y, slice_x]
        radio -= dark
        radio_pos = self._radios_idx_to_pos[radio_idx]
        flat = self.get_flat(radio_pos, dtype=dtype, slice_y=slice_y, slice_x=slice_x)
        flat = flat - dark
        if self.distortion_correction is not None:
            flat = self.distortion_correction.estimate_and_correct(flat, radio)
        radio /= flat
        if self.normalize_srcurrent:
            radio *= self.srcurrent_ratios[radio_pos]
        self.remove_invalid_values(radio)
        return radio


FlatField = FlatFieldArrays


@deprecated_class(
    "FlatFieldDataUrls is deprecated since 2024.2.0 and will be removed in a future version", do_print=True
)
class FlatFieldDataUrls(FlatField):
    def __init__(
        self,
        radios_shape: tuple,
        flats: dict,
        darks: dict,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
        nan_value=1.0,
        radios_srcurrent=None,
        flats_srcurrent=None,
        **chunk_reader_kwargs,
    ):
        """
        Initialize a flat-field normalization process with DataUrls.

        Parameters
        ----------
        radios_shape: tuple
            A tuple describing the shape of the radios stack, in the form
            `(n_radios, n_z, n_x)`.
        flats: dict
            Dictionary where the key is the flat index, and the value is a
            silx.io.DataUrl pointing to the flat.
        darks: dict
            Dictionary where the key is the dark index, and the value is a
            silx.io.DataUrl pointing to the dark.
        radios_indices: array, optional
            Array containing the radios indices. `radios_indices[0]` is the index
            of the first radio, and so on.
        interpolation: str, optional
            Interpolation method for flat-field. See below for more details.
        distortion_correction: DistortionCorrection, optional
            A DistortionCorrection object. If provided, it is used to correct flat distortions based on each radio.
        nan_value: float, optional
            Which float value is used to replace nan/inf after flat-field.


        Other Parameters
        ----------------
        The other named parameters are passed to ChunkReader(). Please read its
        documentation for more information.

        Notes
        ------
        Usually, when doing a scan, only one or a few darks/flats are acquired.
        However, the flat-field normalization has to be performed on each radio,
        although incoming beam can fluctuate between projections.
        The usual way to overcome this is to interpolate between flats.
        If interpolation="nearest", the first flat is used for the first
        radios subset, the second flat is used for the second radios subset,
        and so on.
        If interpolation="linear", the normalization is done as a linear
        function of the radio index.
        """

        flats_arrays_dict = load_images_from_dataurl_dict(flats, **chunk_reader_kwargs)
        darks_arrays_dict = load_images_from_dataurl_dict(darks, **chunk_reader_kwargs)
        super().__init__(
            radios_shape,
            flats_arrays_dict,
            darks_arrays_dict,
            radios_indices=radios_indices,
            interpolation=interpolation,
            distortion_correction=distortion_correction,
            nan_value=nan_value,
            radios_srcurrent=radios_srcurrent,
            flats_srcurrent=flats_srcurrent,
        )



from itertools import product
import tarfile
import os
import numpy as np
from scipy.signal.windows import gaussian
from silx.resources import ExternalResources
from silx.io.dictdump import nxtodict, dicttonx

utilstest = ExternalResources(
    project="nabu", url_base="http://www.silx.org/pub/nabu/data/", env_key="NABU_DATA", timeout=60
)

__big_testdata_dir__ = os.environ.get("NABU_BIGDATA_DIR")
if __big_testdata_dir__ is None or not (os.path.isdir(__big_testdata_dir__)):
    __big_testdata_dir__ = None

__do_long_tests__ = os.environ.get("NABU_LONG_TESTS", False)
if __do_long_tests__:
    try:
        __do_long_tests__ = bool(int(__do_long_tests__))
    except:
        __do_long_tests__ = False

__do_large_mem_tests__ = os.environ.get("NABU_LARGE_MEM_TESTS", False)
if __do_large_mem_tests__:
    try:
        __do_large_mem_tests__ = bool(int(__do_large_mem_tests__))
    except:
        __do_large_mem_tests__ = False


def generate_tests_scenarios(configurations):
    """
    Generate "scenarios" of tests.

    The parameter is a dictionary where:
      - the key is the name of a parameter
      - the value is a list of possible parameters

    This function returns a list of dictionary where:
      - the key is the name of a parameter
      - the value is one value of this parameter
    """
    scenarios = [{key: val for key, val in zip(configurations.keys(), p_)} for p_ in product(*configurations.values())]
    return scenarios


def get_data(*dataset_path):
    """
    Get a dataset file from silx.org/pub/nabu/data
    dataset_args is a list describing a nested folder structures, ex.
    ["path", "to", "my", "dataset.h5"]
    """
    dataset_relpath = os.path.join(*dataset_path)
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    return np.load(dataset_downloaded_path)


def get_array_of_given_shape(img, shape, dtype):
    """
    From a given image, returns an array of the wanted shape and dtype.
    """

    # Tile image until it's big enough.
    # "fun" fact: using any(blabla) crashes but using any([blabla]) does not, because of variables re-evaluation
    while any([i_dim <= s_dim for i_dim, s_dim in zip(img.shape, shape)]):
        img = np.tile(img, (2, 2))
    if len(shape) == 1:
        arr = img[: shape[0], 0]
    elif len(shape) == 2:
        arr = img[: shape[0], : shape[1]]
    else:
        arr = np.tile(img, (shape[0], 1, 1))[: shape[0], : shape[1], : shape[2]]
    return np.ascontiguousarray(np.squeeze(arr), dtype=dtype)


def get_big_data(filename):
    if __big_testdata_dir__ is None:
        return None
    return np.load(os.path.join(__big_testdata_dir__, filename))


def uncompress_file(compressed_file_path, target_directory):
    with tarfile.open(compressed_file_path) as f:
        f.extractall(path=target_directory)


def get_file(fname):
    downloaded_file = dataset_downloaded_path = utilstest.getfile(fname)
    if ".tar" in fname:
        uncompress_file(downloaded_file, os.path.dirname(downloaded_file))
        downloaded_file = downloaded_file.split(".tar")[0]
    return downloaded_file


def compare_arrays(arr1, arr2, tol, diff=None, absolute_value=True, percent=None, method="max", return_residual=False):
    """
    Utility to compare two arrays.

    Parameters
    ----------
    arr1: numpy.ndarray
        First array to compare
    arr2: numpy.ndarray
        Second array to compare
    tol: float
        Tolerance indicating whether arrays are close to eachother.
    diff: numpy.ndarray, optional
        Difference `arr1 - arr2`. If provided, this array is taken instead of `arr1`
        and `arr2`.
    absolute_value: bool, optional
        Whether to take absolute value of the difference.
    percent: float
        If set, a "relative" comparison is performed instead of a subtraction:
        `red(|arr1 - arr2|) / (red(|arr1|) * percent) < tol`
        where "red" is the reduction method (mean, max or median).
    method:
        Reduction method. Can be "max", "mean", or "median".

    Returns
    --------
    (is_close, residual) if return_residual is set to True
    is_close otherwise

    Examples
    --------
    When using method="mean" and absolute_value=True, this function computes
    the Mean Absolute Difference (MAD) metric.
    When also using percent=1.0, this computes the Relative Mean Absolute Difference
    (RMD) metric.
    """
    reductions = {
        "max": np.max,
        "mean": np.mean,
        "median": np.median,
    }
    if method not in reductions:
        raise ValueError("reduction method should be in %s" % str(list(reductions.keys())))
    if diff is None:
        diff = arr1 - arr2
    if absolute_value is not None:
        diff = np.abs(diff)
    residual = reductions[method](diff)
    if percent is not None:
        a1 = np.abs(arr1) if absolute_value else arr1
        residual /= reductions[method](a1)

    res = residual < tol
    if return_residual:
        res = res, residual
    return res


def gaussian_apodization_window(shape, fwhm_ratio=0.7):
    fwhm = fwhm_ratio * np.array(shape)
    sigma = fwhm / 2.355
    return np.outer(*[gaussian(n, s) for n, s in zip(shape, sigma)])


def compare_shifted_images(img1, img2, fwhm_ratio=0.7, return_upper_bound=False):
    """
    Compare two images that are slightly shifted from one another.
    Typically, tomography reconstruction wight slightly different CoR.
    Each image is Fourier-transformed, and the modulus is taken to get rid of the shift between the images.
    An apodization is done to filter the high frequencies that are usually less relevant.

    Parameters
    ----------
    img1: numpy.ndarray
        First image
    img2: numpy.ndarray
        Second image
    fwhm_ratio: float, optional
        Ratio defining the apodization in the frequency domain.
        A small value (eg. 0.2) means that essentually only the low frequencies will be compared.
        A value of 1.0 means no apodization
    return_upper_bound: bool, optional
        Whether to return a (coarse) upper bound of the comparison metric

    Notes
    -----
    This function roughly computes
        |phi(F(img1)) - phi(F(img2))|
    where F is the absolute value of the Fourier transform, and phi some shrinking function (here arcsinh).
    """

    def _fourier_transform(img):
        return np.arcsinh(np.abs(np.fft.fftshift(np.fft.fft2(img))))

    diff = _fourier_transform(img1) - _fourier_transform(img2)
    diff *= gaussian_apodization_window(img1.shape, fwhm_ratio=fwhm_ratio)
    res = np.abs(diff).max()
    if return_upper_bound:
        data_range = np.max(np.abs(img1))
        return res, np.arcsinh(np.prod(img1.shape) * data_range)
    else:
        return res


# To be improved
def generate_nx_dataset(out_fname, image_key, data_volume=None, rotation_angle=None):
    nx_template_file = get_file("dummy.nx.tar.gz")
    nx_dict = nxtodict(nx_template_file)
    nx_entry = nx_dict["entry"]

    def _get_field(dict_, path):
        if path.startswith("/"):
            path = path[1:]
        if path.endswith("/"):
            path = path[:-1]
        split_path = path.split("/")
        if len(split_path) == 1:
            return dict_[split_path[0]]
        return _get_field(dict_[split_path[0]], "/".join(split_path[1:]))

    for name in ["image_key", "image_key_control"]:
        nx_entry["data"][name] = image_key
        nx_entry["instrument"]["detector"][name] = image_key

    if rotation_angle is not None:
        nx_entry["data"]["rotation_angle"] = rotation_angle
        nx_entry["sample"]["rotation_angle"] = rotation_angle

    dicttonx(nx_dict, out_fname)



from math import pi
import numpy as np
from scipy.fft import rfft, irfft
from silx.image.tomography import compute_fourier_filter, get_next_power
from ..processing.padding_base import PaddingBase
from ..utils import check_supported, get_num_threads

# # COMPAT.
# from .filtering_cuda import CudaSinoFilter

# SinoFilter = deprecated_class(
#     "From version 2023.1, 'filtering_cuda.CudaSinoFilter' should be used instead of 'filtering.SinoFilter'. In the future, 'filtering.SinoFilter' will be a numpy-only class.",
#     do_print=True,
# )(CudaSinoFilter)
# #


class SinoFilter:
    available_filters = [
        "ramlak",
        "shepp-logan",
        "cosine",
        "hamming",
        "hann",
        "tukey",
        "lanczos",
        "hilbert",
    ]

    """
    A class for sinogram filtering.
    It does the following:
      - pad input array
      - Fourier transform each row
      - multiply with a 1D or 2D filter
      - inverse Fourier transform
    """

    available_padding_modes = PaddingBase.supported_modes
    default_extra_options = {"cutoff": 1.0, "fft_threads": 0}  # use all threads by default

    def __init__(
        self,
        sino_shape,
        filter_name=None,
        padding_mode="zeros",
        extra_options=None,
    ):
        self._init_extra_options(extra_options)
        self._set_padding_mode(padding_mode)
        self._calculate_shapes(sino_shape)
        self._init_fft()
        self._allocate_memory()
        self._compute_filter(filter_name)

    def _init_extra_options(self, extra_options):
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def _set_padding_mode(self, padding_mode):
        # Compat.
        if padding_mode == "edges":
            padding_mode = "edge"
        if padding_mode == "zeros":
            padding_mode = "constant"
        #
        check_supported(padding_mode, self.available_padding_modes, "padding mode")
        self.padding_mode = padding_mode

    def _calculate_shapes(self, sino_shape):
        self.ndim = len(sino_shape)
        if self.ndim == 2:
            n_angles, dwidth = sino_shape
            n_sinos = 1
        elif self.ndim == 3:
            n_sinos, n_angles, dwidth = sino_shape
        else:
            raise ValueError("Invalid sinogram number of dimensions")
        self.sino_shape = sino_shape
        self.n_angles = n_angles
        self.dwidth = dwidth
        # Make sure to use int() here, otherwise pycuda/pyopencl will crash in some cases
        self.dwidth_padded = int(get_next_power(2 * self.dwidth))
        self.sino_padded_shape = (n_angles, self.dwidth_padded)
        if self.ndim == 3:
            self.sino_padded_shape = (n_sinos,) + self.sino_padded_shape
        sino_f_shape = list(self.sino_padded_shape)
        sino_f_shape[-1] = sino_f_shape[-1] // 2 + 1
        self.sino_f_shape = tuple(sino_f_shape)
        self.pad_left = (self.dwidth_padded - self.dwidth) // 2
        self.pad_right = self.dwidth_padded - self.dwidth - self.pad_left

    def _init_fft(self):
        pass

    def _allocate_memory(self):
        pass

    def set_filter(self, h_filt, normalize=True):
        """
        Set a filter for sinogram filtering.

        Parameters
        ----------
        h_filt: numpy.ndarray
            Array containing the filter. Each line of the sinogram will be filtered with
            this filter. It has to be the Real-to-Complex Fourier Transform
            of some real filter, padded to 2*sinogram_width.
        normalize: bool or float, optional
            Whether to normalize (multiply) the filter with pi/num_angles.
        """
        if h_filt.size != self.sino_f_shape[-1]:
            raise ValueError(
                """
                Invalid filter size: expected %d, got %d.
                Please check that the filter is the Fourier R2C transform of
                some real 1D filter.
                """
                % (self.sino_f_shape[-1], h_filt.size)
            )
        if not (np.iscomplexobj(h_filt)):
            print("Warning: expected a complex Fourier filter")
        self.filter_f = h_filt.copy()
        if normalize:
            self.filter_f *= pi / self.n_angles
        self.filter_f = self.filter_f.astype(np.complex64)

    def _compute_filter(self, filter_name):
        self.filter_name = filter_name or "ram-lak"
        # TODO add this one into silx
        if self.filter_name == "hilbert":
            freqs = np.fft.fftfreq(self.dwidth_padded)
            filter_f = 1.0 / (2 * pi * 1j) * np.sign(freqs)
        #
        else:
            filter_f = compute_fourier_filter(
                self.dwidth_padded,
                self.filter_name,
                cutoff=self.extra_options["cutoff"],
            )
        filter_f = filter_f[: self.dwidth_padded // 2 + 1]  # R2C
        self.set_filter(filter_f, normalize=True)

    def _check_array(self, arr):
        if arr.dtype != np.float32:
            raise ValueError("Expected data type = numpy.float32")
        if arr.shape != self.sino_shape:
            raise ValueError("Expected sinogram shape %s, got %s" % (self.sino_shape, arr.shape))

    def filter_sino(self, sino, output=None, no_output=False):
        """
        Perform the sinogram siltering.

        Parameters
        ----------
        sino: numpy.ndarray or pycuda.gpuarray.GPUArray
            Input sinogram (2D or 3D)
        output: numpy.ndarray or pycuda.gpuarray.GPUArray, optional
            Output array.
        no_output: bool, optional
            If set to True, no copy is be done. The resulting data lies
            in self.d_sino_padded.
        """
        self._check_array(sino)
        # sino_padded = np.pad(
        #     sino, ((0, 0), (0, self.dwidth_padded - self.dwidth)), mode=self.padding_mode
        # )  # pad with a FFT-friendly layout
        sino_padded = np.pad(sino, ((0, 0), (self.pad_left, self.pad_right)), mode=self.padding_mode)
        sino_padded_f = rfft(sino_padded, axis=1, workers=get_num_threads(self.extra_options["fft_threads"]))
        sino_padded_f *= self.filter_f
        sino_filtered = irfft(sino_padded_f, axis=1, workers=get_num_threads(self.extra_options["fft_threads"]))
        if output is None:
            res = np.zeros(self.sino_shape, dtype=np.float32)
        else:
            res = output
        if self.ndim == 2:
            # res[:] = sino_filtered[:, : self.dwidth]  # pylint: disable=E1126 # ?!
            res[:] = sino_filtered[:, self.pad_left : -self.pad_right]  # pylint: disable=E1126 # ?!
        else:
            # res[:] = sino_filtered[:, :, : self.dwidth]  # pylint: disable=E1126 # ?!
            res[:] = sino_filtered[:, :, self.pad_left : -self.pad_right]  # pylint: disable=E1126 # ?!
        return res

    __call__ = filter_sino


def filter_sinogram(
    sinogram,
    padded_width,
    filter_name="ramlak",
    padding_mode="constant",
    normalize=True,
    filter_cutoff=1.0,
    **padding_kwargs,
):
    """
    Simple function to filter sinogram.

    Parameters
    ----------
    sinogram: numpy.ndarray
        Sinogram, two dimensional array with shape (n_angles, sino_width)
    padded_width: int
        Width to use for padding. Must be greater than sinogram width (i.e than sinogram.shape[-1])
    filter_name: str, optional
        Which filter to use. Default is ramlak (roughly equivalent to abs(nu) in frequency domain)
    padding_mode: str, optional
        Which padding mode to use. Default is zero-padding.
    normalize: bool, optional
        Whether to multiply the filtered sinogram with pi/n_angles
    filter_cutoff: float, optional
        frequency cutoff for filter
    """
    n_angles, width = sinogram.shape

    # Initially, padding was done this way
    # sinogram_padded = np.pad(sinogram, ((0, 0), (0, padded_width - width)), mode=padding_mode, **padding_kwargs)

    #
    pad_left = (padded_width - width) // 2
    pad_right = padded_width - width - pad_left
    sinogram_padded = np.pad(sinogram, ((0, 0), (pad_left, pad_right)), mode=padding_mode, **padding_kwargs)
    #

    fourier_filter = compute_fourier_filter(padded_width, filter_name, cutoff=filter_cutoff)
    if normalize:
        fourier_filter *= np.pi / n_angles
    fourier_filter = fourier_filter[: padded_width // 2 + 1]  # R2C
    sino_f = rfft(sinogram_padded, axis=1)
    sino_f *= fourier_filter
    # sino_filtered = irfft(sino_f, axis=1)[:, :width]  # pylint: disable=E1126 # ?!
    sino_filtered = irfft(sino_f, axis=1)[:, pad_left:-pad_right]  # pylint: disable=E1126 # ?!
    return sino_filtered



from .nabu_config import nabu_config, renamed_keys
from .dataset_validator import HelicalDatasetValidator
from ..fullfield import processconfig as ff_processconfig
from ...resources import dataset_analyzer


class ProcessConfig(ff_processconfig.ProcessConfig):
    default_nabu_config = nabu_config
    config_renamed_keys = renamed_keys
    _use_horizontal_translations = False

    def _configure_save_steps(self):
        self._dump_sinogram = False
        steps_to_save = self.nabu_config["pipeline"]["save_steps"]
        if steps_to_save in (None, ""):
            self.steps_to_save = []
            return
        steps_to_save = [s.strip() for s in steps_to_save.split(",")]
        for step in self.processing_steps:
            step = step.strip()
            if step in steps_to_save:
                self.processing_options[step]["save"] = True
                self.processing_options[step]["save_steps_file"] = self.get_save_steps_file(step_name=step)
        # "sinogram" is a special keyword, not explicitly in the processing steps
        if "sinogram" in steps_to_save:
            self._dump_sinogram = True
            self._dump_sinogram_file = self.get_save_steps_file(step_name="sinogram")
        self.steps_to_save = steps_to_save

    def _update_dataset_info_with_user_config(self):
        super()._update_dataset_info_with_user_config()

        self._get_translation_file("reconstruction", "z_per_proj_file", "z_per_proj", last_dim=1)
        self._get_translation_file("reconstruction", "x_per_proj_file", "x_per_proj", last_dim=1)

    def _get_user_sino_normalization(self):
        """is called by the base class but it is not used in helical"""
        pass

    def _coupled_validation(self):
        self.logger.debug("Doing coupled validation")
        self._dataset_validator = HelicalDatasetValidator(self.nabu_config, self.dataset_info)
        for what in ["rec_params", "rec_region", "binning", "subsampling_factor"]:
            setattr(self, what, getattr(self._dataset_validator, what))
            print(what, self._dataset_validator)

    def _browse_dataset(self, dataset_info):
        """
        Browse a dataset and builds a data structure with the relevant information.
        """
        self.logger.debug("Browsing dataset")
        if dataset_info is not None:
            self.dataset_info = dataset_info
        else:
            self.dataset_info = dataset_analyzer.analyze_dataset(
                self.nabu_config["dataset"]["location"],
                extra_options={
                    "exclude_projections": self.nabu_config["dataset"]["exclude_projections"],
                    "hdf5_entry": self.nabu_config["dataset"]["hdf5_entry"],
                },
                logger=self.logger,
            )



import os
from silx.image.phantomgenerator import PhantomGenerator
from scipy.ndimage import shift as scipy_shift
import numpy
import pytest
from nabu.stitching.config import PreProcessedZStitchingConfiguration
from nabu.stitching.config import KEY_IMG_REG_METHOD
from nabu.stitching.overlap import ImageStichOverlapKernel, OverlapStitchingStrategy
from nabu.stitching.z_stitching import (
    PreProcessZStitcher,
)
from nabu.stitching.stitcher_2D import stitch_raw_frames, get_overlap_areas
from nxtomo.nxobject.nxdetector import ImageKey
from nxtomo.utils.transformation import DetYFlipTransformation, DetZFlipTransformation
from nxtomo.application.nxtomo import NXtomo
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from nabu.stitching.utils import ShiftAlgorithm
import h5py


_stitching_configurations = (
    # simple case where shifts are provided
    {
        "n_proj": 4,
        "raw_pos": ((0, 0, 0), (-90, 0, 0), (-180, 0, 0)),  # requested shift to
        "input_pos": ((0, 0, 0), (-90, 0, 0), (-180, 0, 0)),  # requested shift to
        "raw_shifts": ((0, 0), (-90, 0), (-180, 0)),
    },
    # simple case where shift is found from z position
    {
        "n_proj": 4,
        "raw_pos": ((90, 0, 0), (0, 0, 0), (-90, 0, 0)),
        "input_pos": ((90, 0, 0), (0, 0, 0), (-90, 0, 0)),
        "check_bb": ((40, 140), (-50, 50), (-140, -40)),
        "axis_0_params": {
            KEY_IMG_REG_METHOD: ShiftAlgorithm.NONE,
        },
        "axis_2_params": {
            KEY_IMG_REG_METHOD: ShiftAlgorithm.NONE,
        },
        "raw_shifts": ((0, 0), (-90, 0), (-180, 0)),
    },
)


@pytest.mark.parametrize("configuration", _stitching_configurations)
@pytest.mark.parametrize("dtype", (numpy.float32, numpy.int16))
def test_PreProcessZStitcher(tmp_path, dtype, configuration):
    """
    test PreProcessZStitcher class and insure a full stitching can be done automatically.
    """
    n_proj = configuration["n_proj"]
    ref_frame_width = 280
    raw_frame_height = 100
    ref_frame = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(dtype) * 256.0

    # add some mark for image registration
    ref_frame[:, 96] = -3.2
    ref_frame[:, 125] = 9.1
    ref_frame[:, 165] = 4.4
    ref_frame[:, 200] = -2.5
    # create raw data
    frame_0_shift, frame_1_shift, frame_2_shift = configuration["raw_shifts"]
    frame_0 = scipy_shift(ref_frame, shift=frame_0_shift)[:raw_frame_height]
    frame_1 = scipy_shift(ref_frame, shift=frame_1_shift)[:raw_frame_height]
    frame_2 = scipy_shift(ref_frame, shift=frame_2_shift)[:raw_frame_height]

    frames = frame_0, frame_1, frame_2
    frame_0_input_pos, frame_1_input_pos, frame_2_input_pos = configuration["input_pos"]
    frame_0_raw_pos, frame_1_raw_pos, frame_2_raw_pos = configuration["raw_pos"]

    # create a Nxtomo for each of those raw data
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()
    z_position = (
        frame_0_raw_pos[0],
        frame_1_raw_pos[0],
        frame_2_raw_pos[0],
    )
    scans = []
    for (i_frame, frame), z_pos in zip(enumerate(frames), z_position):
        nx_tomo = NXtomo()
        nx_tomo.sample.z_translation = [z_pos] * n_proj
        nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=n_proj, endpoint=False)
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * n_proj
        nx_tomo.instrument.detector.x_pixel_size = 1.0
        nx_tomo.instrument.detector.y_pixel_size = 1.0
        nx_tomo.instrument.detector.distance = 2.3
        nx_tomo.energy = 19.2
        nx_tomo.instrument.detector.data = numpy.asarray([frame] * n_proj)

        file_path = os.path.join(raw_data_dir, f"nxtomo_{i_frame}.nx")
        entry = f"entry000{i_frame}"
        nx_tomo.save(file_path=file_path, data_path=entry)
        scans.append(NXtomoScan(scan=file_path, entry=entry))

    # if requested: check bounding box
    check_bb = configuration.get("check_bb", None)
    if check_bb is not None:
        for scan, expected_bb in zip(scans, check_bb):
            assert scan.get_bounding_box(axis="z") == expected_bb
    output_file_path = os.path.join(output_dir, "stitched.nx")
    output_data_path = "stitched"
    z_stich_config = PreProcessedZStitchingConfiguration(
        stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
        overwrite_results=True,
        axis_0_pos_px=(
            frame_0_input_pos[0],
            frame_1_input_pos[0],
            frame_2_input_pos[0],
        ),
        axis_1_pos_px=(
            frame_0_input_pos[1],
            frame_1_input_pos[1],
            frame_2_input_pos[1],
        ),
        axis_2_pos_px=(
            frame_0_input_pos[2],
            frame_1_input_pos[2],
            frame_2_input_pos[2],
        ),
        axis_0_pos_mm=None,
        axis_1_pos_mm=None,
        axis_2_pos_mm=None,
        input_scans=scans,
        output_file_path=output_file_path,
        output_data_path=output_data_path,
        axis_0_params=configuration.get("axis_0_params", {}),
        axis_1_params=configuration.get("axis_1_params", {}),
        axis_2_params=configuration.get("axis_2_params", {}),
        output_nexus_version=None,
        slices=None,
        slurm_config=None,
        slice_for_cross_correlation="middle",
        pixel_size=None,
    )
    stitcher = PreProcessZStitcher(z_stich_config)
    output_identifier = stitcher.stitch()
    assert output_identifier.file_path == output_file_path
    assert output_identifier.data_path == output_data_path

    created_nx_tomo = NXtomo().load(
        file_path=output_identifier.file_path,
        data_path=output_identifier.data_path,
        detector_data_as="as_numpy_array",
    )

    assert created_nx_tomo.instrument.detector.data.ndim == 3
    mean_abs_error = configuration.get("mean_abs_error", None)
    if mean_abs_error is not None:
        assert (
            numpy.mean(numpy.abs(ref_frame - created_nx_tomo.instrument.detector.data[0, :ref_frame_width, :]))
            < mean_abs_error
        )
    else:
        numpy.testing.assert_array_almost_equal(
            ref_frame, created_nx_tomo.instrument.detector.data[0, :ref_frame_width, :]
        )

    # check also other metadata are here
    assert created_nx_tomo.instrument.detector.distance.value == 2.3
    assert created_nx_tomo.energy.value == 19.2
    numpy.testing.assert_array_equal(
        created_nx_tomo.instrument.detector.image_key_control,
        numpy.asarray([ImageKey.PROJECTION.PROJECTION] * n_proj),
    )

    # check configuration has been saved
    with h5py.File(output_identifier.file_path, mode="r") as h5f:
        assert "stitching_configuration" in h5f[output_identifier.data_path]


slices_to_test_pre = (
    {
        "slices": (None,),
        "complete": True,
    },
    {
        "slices": (("first",), ("middle",), ("last",)),
        "complete": False,
    },
    {
        "slices": ((0, 1, 2), slice(3, -1, 1)),
        "complete": True,
    },
)


def build_nxtomos(output_dir) -> tuple:
    r"""
    build two nxtomos in output_dir and return the list of NXtomos ready to be stitched
      /\
      |        ______________
      |       |  nxtomo 1    |
    Z |       |  frame       |
      |       |~~~~~~~~~~~~~~|
      |       |~~~~~~~~~~~~~~|
      |       |______________|
      |        ______________
      |       |~~~~~~~~~~~~~~|
      |       |~~~~~~~~~~~~~~|
      |       |  nxtomo 2    |
      |       |  frame       |
      |       |______________|
      |
    <-----------------------------------------------
                    y (in acquisition space)

    * ~: represent the overlap area
    """
    n_projs = 100
    raw_data = numpy.arange(100 * 128 * 128).reshape((100, 128, 128))

    # create raw data
    frame_0 = raw_data[:, 60:]
    assert frame_0.ndim == 3
    frame_0_pos = 40
    frame_1 = raw_data[:, 0:80]
    assert frame_1.ndim == 3
    frame_1_pos = 94
    frames = (frame_0, frame_1)
    z_positions = (frame_0_pos, frame_1_pos)

    # create a Nxtomo for each of those raw data

    scans = []
    for (i_frame, frame), z_pos in zip(enumerate(frames), z_positions):
        nx_tomo = NXtomo()
        nx_tomo.sample.z_translation = [z_pos] * n_projs
        nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=n_projs, endpoint=False)
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * n_projs
        nx_tomo.instrument.detector.x_pixel_size = 1.0
        nx_tomo.instrument.detector.y_pixel_size = 1.0
        nx_tomo.instrument.detector.distance = 2.3
        nx_tomo.energy = 19.2
        nx_tomo.instrument.detector.data = frame

        file_path = os.path.join(output_dir, f"nxtomo_{i_frame}.nx")
        entry = f"entry000{i_frame}"
        nx_tomo.save(file_path=file_path, data_path=entry)
        scans.append(NXtomoScan(scan=file_path, entry=entry))
    return scans, z_positions, raw_data


@pytest.mark.parametrize("configuration_dist", slices_to_test_pre)
def test_DistributePreProcessZStitcher(tmp_path, configuration_dist):
    slices = configuration_dist["slices"]
    complete = configuration_dist["complete"]

    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()

    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()

    scans, z_positions, raw_data = build_nxtomos(output_dir=raw_data_dir)
    stitched_nx_tomo = []
    for s in slices:
        output_file_path = os.path.join(output_dir, "stitched_section.nx")
        output_data_path = f"stitched_{s}"
        z_stich_config = PreProcessedZStitchingConfiguration(
            axis_0_pos_px=z_positions,
            axis_1_pos_px=(0, 0),
            axis_2_pos_px=None,
            axis_0_pos_mm=None,
            axis_1_pos_mm=None,
            axis_2_pos_mm=None,
            axis_0_params={},
            axis_1_params={},
            axis_2_params={},
            stitching_strategy=OverlapStitchingStrategy.CLOSEST,
            overwrite_results=True,
            input_scans=scans,
            output_file_path=output_file_path,
            output_data_path=output_data_path,
            output_nexus_version=None,
            slices=s,
            slurm_config=None,
            slice_for_cross_correlation="middle",
            pixel_size=None,
        )
        stitcher = PreProcessZStitcher(z_stich_config)
        output_identifier = stitcher.stitch()
        assert output_identifier.file_path == output_file_path
        assert output_identifier.data_path == output_data_path

        created_nx_tomo = NXtomo().load(
            file_path=output_identifier.file_path,
            data_path=output_identifier.data_path,
            detector_data_as="as_numpy_array",
        )
        stitched_nx_tomo.append(created_nx_tomo)
    assert len(stitched_nx_tomo) == len(slices)
    final_nx_tomo = NXtomo.concatenate(stitched_nx_tomo)
    assert isinstance(final_nx_tomo.instrument.detector.data, numpy.ndarray)
    final_nx_tomo.save(
        file_path=os.path.join(output_dir, "final_stitched.nx"),
        data_path="entry0000",
    )

    if complete:
        len(final_nx_tomo.instrument.detector.data) == 128
        # test middle
        numpy.testing.assert_array_almost_equal(raw_data[1], final_nx_tomo.instrument.detector.data[1, :, :])
    else:
        len(final_nx_tomo.instrument.detector.data) == 3
        # test middle
        numpy.testing.assert_array_almost_equal(raw_data[49], final_nx_tomo.instrument.detector.data[1, :, :])
    # in the case of first, middle and last frames
    # test first
    numpy.testing.assert_array_almost_equal(raw_data[0], final_nx_tomo.instrument.detector.data[0, :, :])

    # test last
    numpy.testing.assert_array_almost_equal(raw_data[-1], final_nx_tomo.instrument.detector.data[-1, :, :])


def test_get_overlap_areas():
    """test get_overlap_areas function"""
    f_upper = numpy.linspace(7, 15, num=9, endpoint=True)
    f_lower = numpy.linspace(0, 12, num=13, endpoint=True)

    o_1, o_2 = get_overlap_areas(
        upper_frame=f_upper,
        lower_frame=f_lower,
        upper_frame_key_line=3,
        lower_frame_key_line=10,
        overlap_size=4,
        stitching_axis=0,
    )

    numpy.testing.assert_array_equal(o_1, o_2)
    numpy.testing.assert_array_equal(o_1, numpy.linspace(8, 11, num=4, endpoint=True))


def test_frame_flip(tmp_path):
    """check it with some NXtomo flipped"""
    ref_frame_width = 280
    n_proj = 10
    raw_frame_width = 100
    ref_frame = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(numpy.float32) * 256.0
    # create raw data
    frame_0_shift = (0, 0)
    frame_1_shift = (-90, 0)
    frame_2_shift = (-180, 0)

    frame_0 = scipy_shift(ref_frame, shift=frame_0_shift)[:raw_frame_width]
    frame_1 = scipy_shift(ref_frame, shift=frame_1_shift)[:raw_frame_width]
    frame_2 = scipy_shift(ref_frame, shift=frame_2_shift)[:raw_frame_width]
    frames = frame_0, frame_1, frame_2

    x_flips = [False, True, True]
    y_flips = [False, False, True]

    def apply_flip(args):
        frame, flip_x, flip_y = args
        if flip_x:
            frame = numpy.fliplr(frame)
        if flip_y:
            frame = numpy.flipud(frame)
        return frame

    frames = map(apply_flip, zip(frames, x_flips, y_flips))

    # create a Nxtomo for each of those raw data
    raw_data_dir = tmp_path / "raw_data"
    raw_data_dir.mkdir()
    output_dir = tmp_path / "output_dir"
    output_dir.mkdir()
    z_position = (90, 0, -90)

    scans = []
    for (i_frame, frame), z_pos, x_flip, y_flip in zip(enumerate(frames), z_position, x_flips, y_flips):
        nx_tomo = NXtomo()
        nx_tomo.sample.z_translation = [z_pos] * n_proj
        nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=n_proj, endpoint=False)
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * n_proj
        nx_tomo.instrument.detector.x_pixel_size = 1.0
        nx_tomo.instrument.detector.y_pixel_size = 1.0
        nx_tomo.instrument.detector.distance = 2.3
        nx_tomo.instrument.detector.transformations.add_transformation(DetZFlipTransformation(flip=x_flip))
        nx_tomo.instrument.detector.transformations.add_transformation(DetYFlipTransformation(flip=y_flip))
        nx_tomo.energy = 19.2
        nx_tomo.instrument.detector.data = numpy.asarray([frame] * n_proj)

        file_path = os.path.join(raw_data_dir, f"nxtomo_{i_frame}.nx")
        entry = f"entry000{i_frame}"
        nx_tomo.save(file_path=file_path, data_path=entry)
        scans.append(NXtomoScan(scan=file_path, entry=entry))

    output_file_path = os.path.join(output_dir, "stitched.nx")
    output_data_path = "stitched"
    assert len(scans) == 3
    z_stich_config = PreProcessedZStitchingConfiguration(
        axis_0_pos_px=(0, -90, -180),
        axis_1_pos_px=(0, 0, 0),
        axis_2_pos_px=None,
        axis_0_pos_mm=None,
        axis_1_pos_mm=None,
        axis_2_pos_mm=None,
        axis_0_params={},
        axis_1_params={},
        axis_2_params={},
        stitching_strategy=OverlapStitchingStrategy.LINEAR_WEIGHTS,
        overwrite_results=True,
        input_scans=scans,
        output_file_path=output_file_path,
        output_data_path=output_data_path,
        output_nexus_version=None,
        slices=None,
        slurm_config=None,
        slice_for_cross_correlation="middle",
        pixel_size=None,
    )
    stitcher = PreProcessZStitcher(z_stich_config)
    output_identifier = stitcher.stitch()
    assert output_identifier.file_path == output_file_path
    assert output_identifier.data_path == output_data_path

    created_nx_tomo = NXtomo().load(
        file_path=output_identifier.file_path,
        data_path=output_identifier.data_path,
        detector_data_as="as_numpy_array",
    )

    assert created_nx_tomo.instrument.detector.data.ndim == 3
    # insure flipping has been taking into account
    numpy.testing.assert_array_almost_equal(ref_frame, created_nx_tomo.instrument.detector.data[0, :ref_frame_width, :])

    assert len(created_nx_tomo.instrument.detector.transformations) == 0



from os import path
from tempfile import TemporaryDirectory
import pytest
import numpy as np
from nabu.io.writer import NXProcessWriter
from nabu.io.reader import import_h5_to_dict
from nabu.testutils import get_data


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls._tmpdir = TemporaryDirectory(prefix="nabu_")
    cls.tempdir = cls._tmpdir.name
    cls.sino_data = get_data("mri_sino500.npz")["data"].astype(np.uint16)
    cls.data = cls.sino_data
    yield
    # teardown
    cls._tmpdir.cleanup()


@pytest.fixture(scope="class")
def bootstrap_h5(request):
    cls = request.cls
    cls._tmpdir = TemporaryDirectory(prefix="nabu_")
    cls.tempdir = cls._tmpdir.name
    cls.data = get_data("mri_sino500.npz")["data"].astype(np.uint16)
    cls.h5_config = {
        "key1": "value1",
        "some_int": 1,
        "some_float": 1.0,
        "some_dict": {
            "numpy_array": np.ones((5, 6), dtype="f"),
            "key2": "value2",
        },
    }
    yield
    # teardown
    cls._tmpdir.cleanup()


@pytest.mark.usefixtures("bootstrap_h5")
class TestNXWriter:
    def test_write_simple(self):
        fname = path.join(self.tempdir, "sino500.h5")
        writer = NXProcessWriter(fname, entry="entry0000")
        writer.write(self.data, "test_write_simple")

    def test_write_with_config(self):
        fname = path.join(self.tempdir, "sino500_cfg.h5")
        writer = NXProcessWriter(fname, entry="entry0000")
        writer.write(self.data, "test_write_with_config", config=self.h5_config)

    def test_overwrite(self):
        fname = path.join(self.tempdir, "sino500_overwrite.h5")
        writer = NXProcessWriter(fname, entry="entry0000", overwrite=True)
        writer.write(self.data, "test_overwrite", config=self.h5_config)

        writer2 = NXProcessWriter(fname, entry="entry0001", overwrite=True)
        writer2.write(self.data, "test_overwrite", config=self.h5_config)

        # overwrite entry0000
        writer3 = NXProcessWriter(fname, entry="entry0000", overwrite=True)
        new_data = self.data.copy()
        new_data += 1
        new_config = self.h5_config.copy()
        new_config["key1"] = "modified value"
        writer3.write(new_data, "test_overwrite", config=new_config)

        res = import_h5_to_dict(fname, "/")
        assert "entry0000" in res
        assert "entry0001" in res
        assert np.allclose(res["entry0000"]["test_overwrite"]["results"]["data"], self.data + 1)
        rec_cfg = res["entry0000"]["test_overwrite"]["configuration"]
        assert rec_cfg["key1"] == "modified value"

    def test_no_overwrite(self):
        fname = path.join(self.tempdir, "sino500_no_overwrite.h5")
        writer = NXProcessWriter(fname, entry="entry0000", overwrite=False)
        writer.write(self.data, "test_no_overwrite")

        writer2 = NXProcessWriter(fname, entry="entry0000", overwrite=False)
        with pytest.raises((RuntimeError, OSError)) as ex:
            writer2.write(self.data, "test_no_overwrite")

        message = "Error should have been raised for trying to overwrite, but got the following: %s" % str(ex.value)



import os
from threading import get_ident
from math import ceil
from multiprocessing.pool import ThreadPool
from posixpath import sep as posix_sep, join as posix_join
import numpy as np
from silx.io import get_data
from silx.io.dictdump import h5todict
from tomoscan.io import HDF5File
from .utils import get_compacted_dataslices, convert_dict_values, get_first_hdf5_entry
from ..misc.binning import binning as image_binning
from ..utils import (
    check_supported,
    deprecated,
    deprecated_class,
    deprecation_warning,
    indices_to_slices,
    compacted_views,
    merge_slices,
    subsample_dict,
    get_3D_subregion,
    get_num_threads,
    get_shape_from_sliced_dims,
    get_size_from_sliced_dimension,
    safe_format,
)

try:
    from fabio.edfimage import EdfImage
except ImportError:
    EdfImage = None


class Reader:
    """
    Abstract class for various file readers.
    """

    def __init__(self, sub_region=None):
        """
        Parameters
        ----------
        sub_region: tuple, optional
            Coordinates in the form (start_x, end_x, start_y, end_y), to read
            a subset of each frame. It can be used for Regions of Interest (ROI).
            Indices start at zero !
        """
        self._set_default_parameters(sub_region)

    def _set_default_parameters(self, sub_region):
        self._set_subregion(sub_region)

    def _set_subregion(self, sub_region):
        self.sub_region = sub_region
        if sub_region is not None:
            start_x, end_x, start_y, end_y = sub_region
            self.start_x = start_x
            self.end_x = end_x
            self.start_y = start_y
            self.end_y = end_y
        else:
            self.start_x = 0
            self.end_x = None
            self.start_y = 0
            self.end_y = None

    def get_data(self, data_url):
        """
        Get data from a silx.io.url.DataUrl
        """
        raise ValueError("Base class")

    def release(self):
        """
        Release the file if needed.
        """
        pass


class NPReader(Reader):
    multi_load = True

    def __init__(self, sub_region=None, mmap=True):
        """
        Reader for NPY/NPZ files. Mostly used for internal development.
        Please refer to the documentation of nabu.io.reader.Reader
        """
        super().__init__(sub_region=sub_region)
        self._file_desc = {}
        self._set_mmap(mmap)

    def _set_mmap(self, mmap):
        self.mmap_mode = "r" if mmap else None

    def _open(self, data_url):
        file_path = data_url.file_path()
        file_ext = self._get_file_type(file_path)
        if file_ext == "npz":
            if file_path not in self._file_desc:
                self._file_desc[file_path] = np.load(file_path, mmap_mode=self.mmap_mode)
            data_ref = self._file_desc[file_path][data_url.data_path()]
        else:
            data_ref = np.load(file_path, mmap_mode=self.mmap_mode)
        return data_ref

    @staticmethod
    def _get_file_type(fname):
        if fname.endswith(".npy"):
            return "npy"
        elif fname.endswith(".npz"):
            return "npz"
        else:
            raise ValueError("Not a numpy file: %s" % fname)

    def get_data(self, data_url):
        data_ref = self._open(data_url)
        data_slice = data_url.data_slice()
        if data_slice is None:
            res = data_ref[self.start_y : self.end_y, self.start_x : self.end_x]
        else:
            res = data_ref[data_slice, self.start_y : self.end_y, self.start_x : self.end_x]
        return res

    def release(self):
        for fname, fdesc in self._file_desc.items():
            if fdesc is not None:
                fdesc.close()
                self._file_desc[fname] = None

    def __del__(self):
        self.release()


class EDFReader(Reader):
    multi_load = False  # not implemented

    def __init__(self, sub_region=None):
        """
        A class for reading series of EDF Files.
        Multi-frames EDF are not supported.
        """
        if EdfImage is None:
            raise ImportError("Need fabio to use this reader")
        super().__init__(sub_region=sub_region)
        self._reader = EdfImage()
        self._first_fname = None

    def read(self, fname):
        if self._first_fname is None:
            self._first_fname = fname
            self._reader.read(fname)
        if self.sub_region is None:
            data = self._reader.data
        else:
            data = self._reader.fast_read_roi(fname, (slice(self.start_y, self.end_y), slice(self.start_x, self.end_x)))
        # self._reader.close()
        return data

    def get_data(self, data_url):
        return self.read(data_url.file_path())


class HDF5Reader(Reader):
    multi_load = True

    def __init__(self, sub_region=None):
        """
        A class for reading a HDF5 File.
        """
        super().__init__(sub_region=sub_region)
        self._file_desc = {}

    def _open(self, file_path):
        if file_path not in self._file_desc:
            self._file_desc[file_path] = HDF5File(file_path, "r", swmr=True)

    def get_data(self, data_url):
        file_path = data_url.file_path()
        self._open(file_path)
        h5dataset = self._file_desc[file_path][data_url.data_path()]
        data_slice = data_url.data_slice()
        if data_slice is None:
            res = h5dataset[self.start_y : self.end_y, self.start_x : self.end_x]
        else:
            res = h5dataset[data_slice, self.start_y : self.end_y, self.start_x : self.end_x]

        return res

    def release(self):
        for fname, fdesc in self._file_desc.items():
            if fdesc is not None:
                try:
                    fdesc.close()
                    self._file_desc[fname] = None
                except Exception as exc:
                    print("Error while closing %s: %s" % (fname, str(exc)))

    def __del__(self):
        self.release()


class HDF5Loader:
    """
    An alternative class to HDF5Reader where information is first passed at class instantiation
    """

    def __init__(self, fname, data_path, sub_region=None, data_buffer=None, pre_allocate=True, dtype="f"):
        self.fname = fname
        self.data_path = data_path
        self._set_subregion(sub_region)
        if not ((data_buffer is not None) ^ (pre_allocate is True)):
            raise ValueError("Please provide either 'data_buffer' or 'pre_allocate'")
        self.data = data_buffer
        self._loaded = False
        self.expected_shape = get_hdf5_dataset_shape(fname, data_path, sub_region=sub_region)
        if pre_allocate:
            self.data = np.zeros(self.expected_shape, dtype=dtype)

    def _set_subregion(self, sub_region):
        self.sub_region = sub_region
        if sub_region is not None:
            start_z, end_z, start_y, end_y, start_x, end_x = sub_region
            self.start_x, self.end_x = start_x, end_x
            self.start_y, self.end_y = start_y, end_y
            self.start_z, self.end_z = start_z, end_z
        else:
            self.start_x, self.end_x = None, None
            self.start_y, self.end_y = None, None
            self.start_z, self.end_z = None, None

    def load_data(self, force_load=False, output=None):
        if self._loaded and not force_load:
            return self.data
        output = self.data if output is None else output
        with HDF5File(self.fname, "r") as fdesc:
            if output is None:
                output = fdesc[self.data_path][
                    self.start_z : self.end_z, self.start_y : self.end_y, self.start_x : self.end_x
                ]
            else:
                output[:] = fdesc[self.data_path][
                    self.start_z : self.end_z, self.start_y : self.end_y, self.start_x : self.end_x
                ]
        self._loaded = True
        return output


@deprecated_class("ChunkReader is deprecated since 2024.2.0 and will be removed in a future version", do_print=True)
class ChunkReader:
    """
    A reader of chunk of images.
    """

    def __init__(
        self,
        files,
        sub_region=None,
        detector_corrector=None,
        pre_allocate=True,
        data_buffer=None,
        convert_float=False,
        shape=None,
        dtype=None,
        binning=None,
        dataset_subsampling=None,
        num_threads=None,
    ):
        """
        Initialize a "ChunkReader". A chunk is a stack of images.

        Parameters
        ----------
        files: dict
            Dictionary where the key is the file/data index, and the value is a
            silx.io.url.DataUrl pointing to the data.
            The dict must contain only the files which shall be used !
            Note: the shape and data type is infered from the first data file.
        sub_region: tuple, optional
            If provided, this must be a tuple in the form
            (start_x, end_x, start_y, end_y). Each image will be cropped to this
            region. This is used to specify a chunk of files.
            Each of the parameters can be None, in this case the default start
            and end are taken in each dimension.
        pre_allocate: bool
            Whether to pre-allocate data before reading.
        data_buffer: array-like, optional
            If `pre_allocate` is set to False, this parameter has to be provided.
            It is an array-like object which will hold the data.
        convert_float: bool
            Whether to convert data to float32, regardless of the input data type.
        shape: tuple, optional
            Shape of each image. If not provided, it is inferred from the first image
            in the collection.
        dtype: `numpy.dtype`, optional
            Data type of each image. If not provided, it is inferred from the first image
            in the collection.
        binning: int or tuple of int, optional
            Whether to bin the data. If multi-dimensional binning is done,
            the parameter must be in the form (binning_x, binning_y).
            Each image will be binned by these factors.
        dataset_subsampling: int or tuple, optional
            Subsampling factor when reading the images.
            If an integer `n` is provided, then one image out of `n` will be read.
            If a tuple of integers (step, begin) is given, the data is read as data[begin::step]
        num_threads: int, optional
            Number of threads to use for binning the data.
            Default is to use all available threads.
            This parameter has no effect when binning is disabled.

        Notes
        ------
        The files are provided as a collection of `silx.io.DataURL`. The file type
        is inferred from the extension.

        Binning is different from subsampling. Using binning will not speed up
        the data retrieval (quite the opposite), since the whole (subregion of) data
        is read and then binning is performed.
        """
        self.detector_corrector = detector_corrector
        self._get_reader_class(files)
        self.dataset_subsampling = dataset_subsampling
        self.num_threads = get_num_threads(num_threads)
        self._set_files(files)
        self._get_shape_and_dtype(shape, dtype, binning)
        self._set_subregion(sub_region)
        self._init_reader()
        self._loaded = False
        self.convert_float = convert_float
        if convert_float:
            self.out_dtype = np.float32
        else:
            self.out_dtype = self.dtype
        if not ((data_buffer is not None) ^ (pre_allocate is True)):
            raise ValueError("Please provide either 'data_buffer' or 'pre_allocate'")
        self.files_data = data_buffer
        if data_buffer is not None:
            # overwrite out_dtype
            self.out_dtype = data_buffer.dtype
            if data_buffer.shape != self.chunk_shape:
                raise ValueError("Expected shape %s but got %s" % (self.shape, data_buffer.shape))
        if pre_allocate:
            self.files_data = np.zeros(self.chunk_shape, dtype=self.out_dtype)
        if (self.binning is not None) and (np.dtype(self.out_dtype).kind in ["u", "i"]):
            raise ValueError(
                "Output datatype cannot be integer when using binning. Please set the 'convert_float' parameter to True or specify a 'data_buffer'."
            )

    def _set_files(self, files):
        if len(files) == 0:
            raise ValueError("Expected at least one data file")
        self._files_begin_idx = 0
        if isinstance(self.dataset_subsampling, (tuple, list)):
            self._files_begin_idx = self.dataset_subsampling[1]
            self.dataset_subsampling = self.dataset_subsampling[0]
        self.n_files = len(files)
        self.files = files
        self._sorted_files_indices = sorted(files.keys())
        self._fileindex_to_idx = dict.fromkeys(self._sorted_files_indices)
        self._configure_subsampling()

    def _infer_file_type(self, files):
        fname = files[sorted(files.keys())[0]].file_path()
        ext = os.path.splitext(fname)[-1].replace(".", "")
        if ext not in Readers:
            raise ValueError("Unknown file format %s. Supported formats are: %s" % (ext, str(Readers.keys())))
        return ext

    def _get_reader_class(self, files):
        ext = self._infer_file_type(files)
        reader_class = Readers[ext]
        self._reader_class = reader_class

    def _get_shape_and_dtype(self, shape, dtype, binning):
        if shape is None or dtype is None:
            shape, dtype = self._infer_shape_and_dtype()
        assert len(shape) == 2, "Expected the shape of an image (2-tuple)"
        self.shape_total = shape
        self.dtype = dtype
        self._set_binning(binning)

    def _configure_subsampling(self):
        dataset_subsampling = self.dataset_subsampling
        self.files_subsampled = self.files
        if dataset_subsampling is not None and dataset_subsampling > 1:
            self.files_subsampled = subsample_dict(self.files, dataset_subsampling)
            self.n_files = len(self.files_subsampled)
            if not (self._reader_class.multi_load):
                # 3D loading not supported for this reader.
                # Data is loaded frames by frame, so subsample directly self.files
                self.files = self.files_subsampled
                self._sorted_files_indices = sorted(self.files.keys())
                self._fileindex_to_idx = dict.fromkeys(self._sorted_files_indices)

    def _infer_shape_and_dtype(self):
        self._reader_entire_image = self._reader_class()
        first_file_dataurl = self.files[self._sorted_files_indices[0]]
        first_file_data = self._reader_entire_image.get_data(first_file_dataurl)
        return first_file_data.shape, first_file_data.dtype

    def _set_subregion(self, sub_region):
        sub_region = sub_region or (None, None, None, None)
        start_x, end_x, start_y, end_y = sub_region
        if start_x is None:
            start_x = 0
        if start_y is None:
            start_y = 0
        if end_x is None:
            end_x = self.shape_total[1]
        if end_y is None:
            end_y = self.shape_total[0]
        self.sub_region = (start_x, end_x, start_y, end_y)
        self.shape = (end_y - start_y, end_x - start_x)
        if self.binning is not None:
            self.shape = (self.shape[0] // self.binning[1], self.shape[1] // self.binning[0])
        self.chunk_shape = (self.n_files,) + self.shape
        if self.detector_corrector is not None:
            self.detector_corrector.set_sub_region_transformation(target_sub_region=self.sub_region)

    def _init_reader(self):
        # instantiate reader with user params
        if self.detector_corrector is not None:
            adapted_subregion = self.detector_corrector.get_adapted_subregion(self.sub_region)
        else:
            adapted_subregion = self.sub_region

        self.file_reader = self._reader_class(sub_region=adapted_subregion)

    def _set_binning(self, binning):
        if binning is None:
            self.binning = None
            return
        if np.isscalar(binning):
            binning = (binning, binning)
        else:
            assert len(binning) == 2, "Expected binning in the form (binning_x, binning_y)"
        if binning[0] == 1 and binning[1] == 1:
            self.binning = None
            return
        for b in binning:
            if int(b) != b:
                raise ValueError("Expected an integer number for binning values, but got %s" % binning)
        self.binning = binning

    def get_data(self, file_url):
        """
        Get the data associated to a file url.
        """
        arr = self.file_reader.get_data(file_url)

        if arr.ndim == 2:
            if self.detector_corrector is not None:
                arr = self.detector_corrector.transform(arr)
            if self.binning is not None:
                arr = image_binning(arr, self.binning[::-1])

        else:
            if self.detector_corrector is not None:
                if self.detector_corrector is not None:
                    _, (
                        src_x_start,
                        src_x_end,
                        src_z_start,
                        src_z_end,
                    ) = self.detector_corrector.get_actual_shapes_source_target()

                    arr_target = np.empty([len(arr), src_z_end - src_z_start, src_x_end - src_x_start], "f")

                def apply_corrector(i_img_tuple):
                    i, img = i_img_tuple
                    arr_target[i] = self.detector_corrector.transform(img)

                with ThreadPool(self.num_threads) as tp:
                    tp.map(apply_corrector, enumerate(arr))
                arr = arr_target
            if self.binning is not None:
                nz = arr.shape[0]
                res = np.zeros((nz,) + image_binning(arr[0], self.binning[::-1]).shape, dtype="f")

                def apply_binning(img_res_tuple):
                    img, res = img_res_tuple
                    res[:] = image_binning(img, self.binning[::-1])

                with ThreadPool(self.num_threads) as tp:
                    tp.map(apply_binning, zip(arr, res))
                arr = res

        return arr

    def _load_single(self):
        for i, fileidx in enumerate(self._sorted_files_indices):
            file_url = self.files[fileidx]
            self.files_data[i] = self.get_data(file_url)
            self._fileindex_to_idx[fileidx] = i

    def _load_multi(self):
        urls_compacted = get_compacted_dataslices(
            self.files, subsampling=self.dataset_subsampling, begin=self._files_begin_idx
        )
        loaded = {}
        start_idx = 0
        sorted_files_indices = sorted(urls_compacted.keys())
        for idx in sorted_files_indices:
            url = urls_compacted[idx]
            url_str = str(url)
            is_loaded = loaded.get(url_str, False)
            if is_loaded:
                continue
            ds = url.data_slice()
            delta_z = ds.stop - ds.start
            if ds.step is not None and ds.step > 1:
                delta_z = ceil(delta_z / ds.step)
            end_idx = start_idx + delta_z
            self.files_data[start_idx:end_idx] = self.get_data(url)
            start_idx += delta_z
            loaded[url_str] = True

    def load_files(self, overwrite: bool = False):
        """
        Load the files whose links was provided at class instantiation.

        Parameters
        -----------
        overwrite: bool, optional
            Whether to force reloading the files if already loaded.
        """
        if self._loaded and not (overwrite):
            raise ValueError("Radios were already loaded. Call load_files(overwrite=True) to force reloading")
        if self.file_reader.multi_load:
            self._load_multi()
        else:
            self._load_single()
        self._loaded = True

    load_data = load_files

    @property
    def data(self):
        return self.files_data


class VolReaderBase:
    """
    Base class with common code for data readers (EDFStackReader, NXTomoReader, etc)
    """

    def __init__(self, *args, **kwargs):
        raise ValueError("Base class")

    def _set_subregion(self, sub_region):
        slice_angle, slice_z, slice_x = (None, None, None)
        if isinstance(sub_region, slice):
            # Assume selection is done only along dimension 0
            slice_angle = sub_region
            slice_z = None
            slice_x = None
        if isinstance(sub_region, (tuple, list)):
            slice_angle, slice_z, slice_x = sub_region
        self.sub_region = (slice_angle or slice(None, None), slice_z or slice(None, None), slice_x or slice(None, None))

    def _set_processing_function(self, processing_func, processing_func_args, processing_func_kwargs):
        self.processing_func = processing_func
        self._processing_func_args = processing_func_args or []
        self._processing_func_kwargs = processing_func_kwargs or {}

    def _get_output(self, array):
        if array is not None:
            if array.shape != self.output_shape:
                raise ValueError("Expected output shape %s but got %s" % (self.output_shape, array.shape))
            if array.dtype != self.output_dtype:
                raise ValueError("Expected output dtype '%s' but got '%s'" % (self.output_dtype, array.dtype))
            output = array
        else:
            output = np.zeros(self.output_shape, dtype=self.output_dtype)
        return output

    def get_frames_indices(self):
        return np.arange(self.data_shape_total[0])[self._source_selection[0]]


class NXTomoReader(VolReaderBase):
    image_key_path = "instrument/detector/image_key_control"
    multiple_frames_per_file = True

    def __init__(
        self,
        filename,
        data_path="{entry}/instrument/detector/data",
        sub_region=None,
        image_key=0,
        output_dtype=np.float32,
        processing_func=None,
        processing_func_args=None,
        processing_func_kwargs=None,
    ):
        """
        Read a HDF5 file in NXTomo layout.

        Parameters
        ----------
        filename: str
            Path to the file to read.
        data_path: str
            Path within the HDF5 file, eg. "entry/instrument/detector/data".
            Default is {entry}/data/data where {entry} is a magic keyword for the first entry.
        sub_region: slice or tuple, optional
            Region to select within the data, once the "image key" selection has been done.
            If None, all the data (corresponding to image_key) is selected.
            If slice(start, stop) is provided, the selection is done along dimension 0.
            Otherwise, it must be a 3-tuple of slices in the form
            (slice(start_angle, end_angle, step_angle), slice(start_z, end_z, step_z), slice(start_x, end_x, step_x))
            Each of the parameters can be None, in this case the default start and end are taken in each dimension.
        output_dtype: numpy.dtype, optional
            Output data type if the data memory is allocated by this class. Default is float32.
        image_key: int, or None, optional
            Image type to read (see NXTomo documentation). 0 for projections, 1 for flat-field, 2 for dark field.
            If set to None, all the data will be read.
        processing_func: callable, optional
            Function to be called on each loaded stack of images.
            If provided, this function first argument must be the source buffer (3D array: stack of raw images),
            and the second argument must be the destination buffer (3D array, stack of output images). It can be None.

        Other parameters
        ----------------
        The other parameters are passed to "processing_func" if this parameter is not None.

        """
        self.filename = filename
        self.data_path = safe_format(data_path or "", entry=get_first_hdf5_entry(filename))
        self._set_image_key(image_key)
        self._set_subregion(sub_region)
        self._get_shape()
        self._set_processing_function(processing_func, processing_func_args, processing_func_kwargs)
        self._get_source_selection()
        self._init_output(output_dtype)

    def _get_input_dtype(self):
        return get_hdf5_dataset_dtype(self.filename, self.data_path)

    def _init_output(self, output_dtype):
        output_shape = get_shape_from_sliced_dims(self.data_shape_total, self._source_selection)
        self.output_dtype = output_dtype
        self._output_shape_no_processing = output_shape
        if self.processing_func is not None:
            test_subvolume = np.zeros((1,) + output_shape[1:], dtype=self._get_input_dtype())
            out = self.processing_func(
                test_subvolume, None, *self._processing_func_args, **self._processing_func_kwargs
            )
            output_image_shape = out.shape[1:]
            output_shape = (output_shape[0],) + output_image_shape
        self.output_shape = output_shape
        self._tmp_dst_buffer = None

    def _set_image_key(self, image_key):
        self.image_key = image_key
        entry = get_entry_from_h5_path(self.data_path)
        image_key_path = posix_join(entry, self.image_key_path)
        with HDF5File(self.filename, "r") as f:
            image_key_val = f[image_key_path][()]
        idx = np.where(image_key_val == image_key)[0]
        if len(idx) == 0:
            raise FileNotFoundError("No frames found with image key = %d" % image_key)
        self._image_key_slices = indices_to_slices(idx)

    def _get_shape(self):
        # Shape of the total HDF5-NXTomo data (including darks and flats)
        self.data_shape_total = get_hdf5_dataset_shape(self.filename, self.data_path)
        # Shape of the data once the "image key" is selected
        n_imgs = self.data_shape_total[0]
        self.data_shape_imagekey = (
            sum([get_size_from_sliced_dimension(n_imgs, slice_) for slice_ in self._image_key_slices]),
        ) + self.data_shape_total[1:]

        # Shape of the data after sub-regions are selected
        self.data_shape_subregion = get_shape_from_sliced_dims(self.data_shape_imagekey, self.sub_region)

    def _get_source_selection(self):
        if len(self._image_key_slices) == 1 and self.processing_func is None:
            # Simple case:
            #   - One single chunk to load, i.e len(self._image_key_slices) == 1
            #   - No on-the-fly processing (binning, distortion correction, ...)
            # In this case, we can use h5py read_direct() to avoid extraneous memory consumption
            image_key_slice = self._image_key_slices[0]
            # merge image key selection and user selection (if any)
            self._source_selection = (
                merge_slices(image_key_slice, self.sub_region[0] or slice(None, None)),
            ) + self.sub_region[1:]
        else:
            user_selection_dim0 = self.sub_region[0]
            indices = np.arange(self.data_shape_total[0])
            data_selection_indices_axis0 = np.hstack(
                [indices[image_key_slice][user_selection_dim0] for image_key_slice in self._image_key_slices]
            )
            self._source_selection = (data_selection_indices_axis0,) + self.sub_region[1:]

    def _get_temporary_buffer(self, convert_after_reading):
        if self._tmp_dst_buffer is None:
            shape = self._output_shape_no_processing
            dtype = self.output_dtype if not (convert_after_reading) else self._get_input_dtype()
            self._tmp_dst_buffer = np.zeros(shape, dtype=dtype)
        return self._tmp_dst_buffer

    def load_data(self, output=None, convert_after_reading=True):
        """
        Read data.

        Parameters
        -----------
        output: array-like, optional
            Destination 3D array that will hold the data.
            If provided, use this memory buffer instead of allocating the memory.
            Its shape must be compatible with the selection of 'sub_region' and 'image_key'.
        conver_after_reading: bool, optional
            Whether to do the dtype conversion (if any, eg. uint16 to float32) after data reading.
            With using h5py's read_direct(), reading from uint16 to float32 is extremely slow,
            so data type conversion should be done after reading.
            The drawback is that it requires more memory.
        """
        output = self._get_output(output)
        convert_after_reading &= np.dtype(self.output_dtype) != np.dtype(self._get_input_dtype())
        if convert_after_reading or self.processing_func is not None:
            dst_buffer = self._get_temporary_buffer(convert_after_reading)
        else:
            dst_buffer = output

        with HDF5File(self.filename, "r") as f:
            dptr = f[self.data_path]
            dptr.read_direct(
                dst_buffer,
                source_sel=self._source_selection,
                dest_sel=None,
            )
        if self.processing_func is not None:
            self.processing_func(dst_buffer, output, *self._processing_func_args, **self._processing_func_kwargs)
        elif dst_buffer.ctypes.data != output.ctypes.data:
            output[:] = dst_buffer[:]  # cast

        return output


class NXDarksFlats:

    _reduce_func = {
        "median": np.median,
        "mean": np.mean,
    }

    def __init__(self, filename, **nxtomoreader_kwargs):
        nxtomoreader_kwargs.pop("image_key", None)
        self.darks_reader = NXTomoReader(filename, image_key=2, **nxtomoreader_kwargs)
        self.flats_reader = NXTomoReader(filename, image_key=1, **nxtomoreader_kwargs)
        self._raw_darks = None
        self._raw_flats = None

    def _get_raw_frames(self, what, force_reload=False, as_multiple_array=True):
        check_supported(what, ["darks", "flats"], "frame type")
        loaded_frames = getattr(self, "_raw_%s" % what)
        reader = getattr(self, "%s_reader" % what)
        if force_reload or loaded_frames is None:
            loaded_frames = reader.load_data()
            setattr(self, "_raw_%s" % what, loaded_frames)
        res = loaded_frames
        if as_multiple_array:
            slices_ = compacted_views(reader._image_key_slices)
            return [res[slice_] for slice_ in slices_]
        return res

    def _get_reduced_frames(self, what, method="mean", force_reload=False, as_dict=False):
        raw_frames = self._get_raw_frames(what, force_reload=force_reload, as_multiple_array=True)
        reduced_frames = [self._reduce_func[method](frames, axis=0) for frames in raw_frames]
        reader = getattr(self, "%s_reader" % what)
        if as_dict:
            return {k: v for k, v in zip([s.start for s in reader._image_key_slices], reduced_frames)}
        return reduced_frames

    def get_raw_darks(self, force_reload=False, as_multiple_array=True):
        return self._get_raw_frames("darks", force_reload=force_reload, as_multiple_array=as_multiple_array)

    def get_raw_flats(self, force_reload=False, as_multiple_array=True):
        return self._get_raw_frames("flats", force_reload=force_reload, as_multiple_array=as_multiple_array)

    def get_reduced_darks(self, method="mean", force_reload=False, as_dict=False):
        return self._get_reduced_frames("darks", method=method, force_reload=force_reload, as_dict=as_dict)

    def get_reduced_flats(self, method="median", force_reload=False, as_dict=False):
        return self._get_reduced_frames("flats", method=method, force_reload=force_reload, as_dict=as_dict)

    def get_raw_current(self, h5_path="{entry}/control/data"):
        h5_path = safe_format(h5_path, entry=self.flats_reader.data_path.split(posix_sep)[0])
        with HDF5File(self.flats_reader.filename, "r") as f:
            current = f[h5_path][()]
        return {sl.start: current[sl] for sl in self.flats_reader._image_key_slices}

    def get_reduced_current(self, h5_path="{entry}/control/data", method="median"):
        current = self.get_raw_current(h5_path=h5_path)
        return {k: self._reduce_func[method](current[k]) for k in current.keys()}


class EDFStackReader(VolReaderBase):
    multiple_frames_per_file = False

    def __init__(
        self,
        filenames,
        sub_region=None,
        output_dtype=np.float32,
        n_reading_threads=1,
        processing_func=None,
        processing_func_args=None,
        processing_func_kwargs=None,
    ):
        self.filenames = filenames
        self.n_reading_threads = n_reading_threads
        self._set_subregion(sub_region)
        self._get_shape()
        self._set_processing_function(processing_func, processing_func_args, processing_func_kwargs)
        self._get_source_selection()
        self._init_output(output_dtype)

    def _get_input_dtype(self):
        return EDFReader().read(self.filenames[0]).dtype

    def _get_shape(self):
        first_filename = self.filenames[0]

        # Shape of the total data (without subregion selection)
        reader_all = EDFReader()
        first_frame_full = reader_all.read(first_filename)
        self.data_shape_total = (len(self.filenames),) + first_frame_full.shape
        self.input_dtype = first_frame_full.dtype

        self.data_shape_subregion = get_shape_from_sliced_dims(
            self.data_shape_total, self.sub_region
        )  # might fail if sub_region is not a slice ?

    def _init_output(self, output_dtype):
        output_shape = get_shape_from_sliced_dims(self.data_shape_total, self._source_selection)
        self.output_dtype = output_dtype
        if self.processing_func is not None:
            test_image = np.zeros(output_shape[1:], dtype=self._get_input_dtype())
            out = self.processing_func(test_image, *self._processing_func_args, **self._processing_func_kwargs)
            output_image_shape = out.shape
            output_shape = (output_shape[0],) + output_image_shape
        self.output_shape = output_shape

    def _get_source_selection(self):
        self._source_selection = self.sub_region

        self._sub_region_xy_for_edf_reader = (
            self.sub_region[2].start or 0,
            self.sub_region[2].stop or self.data_shape_total[2],
            self.sub_region[1].start or 0,
            self.sub_region[1].stop or self.data_shape_total[1],
        )
        self.filenames_subsampled = self.filenames[self.sub_region[0]]

    def load_data(self, output=None):
        output = self._get_output(output)

        readers = {}

        def _init_reader_thread():
            readers[get_ident()] = EDFReader(self._sub_region_xy_for_edf_reader)

        def _get_data(i_fname):
            i, fname = i_fname
            reader = readers[get_ident()]
            frame = reader.read(fname)
            if self.processing_func is not None:
                frame = self.processing_func(frame, *self._processing_func_args, **self._processing_func_kwargs)
            output[i] = frame

        with ThreadPool(self.n_reading_threads, initializer=_init_reader_thread) as tp:
            tp.map(
                _get_data,
                zip(
                    range(len(self.filenames_subsampled)),
                    self.filenames_subsampled,
                ),
            )

        return output


Readers = {
    "edf": EDFReader,
    "hdf5": HDF5Reader,
    "h5": HDF5Reader,
    "nx": HDF5Reader,
    "npz": NPReader,
    "npy": NPReader,
}


@deprecated(
    "Function load_images_from_dataurl_dict is deprecated and will be removed in a fugure version", do_print=True
)
def load_images_from_dataurl_dict(data_url_dict, sub_region=None, dtype="f", binning=None):
    """
    Load a dictionary of dataurl into numpy arrays.

    Parameters
    ----------
    data_url_dict: dict
        A dictionary where the keys are integers (the index of each image in the dataset),
        and the values are numpy.ndarray (data_url_dict).
    sub_region: tuple, optional
        Tuple in the form (y_subregion, x_subregion)
        where xy_subregion is a tuple in the form slice(start, stop, step)

    Returns
    --------
    res: dict
        A dictionary where the keys are the same as `data_url_dict`, and the values are numpy arrays.

    Notes
    -----
    This function is used to load flats/darks. Usually, these are reduced flats/darks, meaning that
    'data_url_dict' is a collection of a handful of files (less than 10).
    To load more frames, it would be best to use NXTomoReader / EDFStackReader.
    """
    res = {}
    if sub_region is not None and not isinstance(sub_region[0], slice):
        if len(sub_region) == 4:  # (start_y, end_y, start_x, end_x)
            deprecation_warning(
                "The parameter 'sub_region' was passed as (start_x, end_x, start_y, end_y). This is deprecated and will yield an error in the future. Please use the syntax ((start_z, end_z), (start_x, end_x))",
                do_print=True,
                func_name="load_images_from_dataurl_dict",
            )
            sub_region = (slice(sub_region[2], sub_region[3]), slice(sub_region[0], sub_region[1]))
        else:  # ((start_z, end_z), (start_x, end_x))
            sub_region = tuple(slice(s[0], s[1]) for s in sub_region)
    for idx, data_url in data_url_dict.items():
        frame = get_data(data_url)
        if sub_region is not None:
            frame = frame[sub_region[0], sub_region[1]]
        if binning is not None:
            frame = image_binning(frame, binning, out_dtype=dtype)
        res[idx] = frame
    return res


def load_images_stack_from_hdf5(fname, h5_data_path, sub_region=None):
    """
    Load a 3D dataset from a HDF5 file.

    Parameters
    -----------
    fname: str
        File path
    h5_data_path: str
        Data path within the HDF5 file
    sub_region: tuple, optional
        Tuple indicating which sub-volume to load, in the form (xmin, xmax, ymin, ymax, zmin, zmax)
        where the 3D dataset has the python shape (N_Z, N_Y, N_X).
        This means that the data will be loaded as `data[zmin:zmax, ymin:ymax, xmin:xmax]`.
    """
    xmin, xmax, ymin, ymax, zmin, zmax = get_3D_subregion(sub_region)
    with HDF5File(fname, "r") as f:
        d_ptr = f[h5_data_path]
        data = d_ptr[zmin:zmax, ymin:ymax, xmin:xmax]
    return data


def get_hdf5_dataset_shape(fname, h5_data_path, sub_region=None):
    zmin, zmax, ymin, ymax, xmin, xmax = get_3D_subregion(sub_region)
    with HDF5File(fname, "r") as f:
        d_ptr = f[h5_data_path]
        shape = d_ptr.shape
    n_z, n_y, n_x = shape
    # perhaps there is more elegant
    res_shape = []
    for n, bounds in zip([n_z, n_y, n_x], ((zmin, zmax), (ymin, ymax), (xmin, xmax))):
        res_shape.append(np.arange(n)[bounds[0] : bounds[1]].size)
    return tuple(res_shape)


def get_hdf5_dataset_dtype(fname, h5_data_path):
    with HDF5File(fname, "r") as f:
        d_ptr = f[h5_data_path]
        dtype = d_ptr.dtype
    return dtype


def get_entry_from_h5_path(h5_path):
    v = h5_path.split(posix_sep)
    return v[0] or v[1]


def check_virtual_sources_exist(fname, data_path):
    with HDF5File(fname, "r") as f:
        if data_path not in f:
            print("No dataset %s in file %s" % (data_path, fname))
            return False
        dptr = f[data_path]
        if not dptr.is_virtual:
            return True
        for vsource in dptr.virtual_sources():
            vsource_fname = os.path.join(os.path.dirname(dptr.file.filename), vsource.file_name)
            if not os.path.isfile(vsource_fname):
                print("No such file: %s" % vsource_fname)
                return False
            elif not check_virtual_sources_exist(vsource_fname, vsource.dset_name):
                print("Error with virtual source %s" % vsource_fname)
                return False
    return True


def import_h5_to_dict(h5file, h5path, asarray=False):
    """
    Wrapper on top of silx.io.dictdump.dicttoh5 replacing "None" with None

    Parameters
    -----------
    h5file: str
        File name
    h5path: str
        Path in the HDF5 file
    asarray: bool, optional
        Whether to convert each numeric value to an 0D array. Default is False.
    """
    dic = h5todict(h5file, path=h5path, asarray=asarray)
    modified_dic = convert_dict_values(dic, {"None": None}, bytes_tostring=True)
    return modified_dic



import numpy as np
from nabu.utils import get_cuda_srcfile, updiv
from .muladd import MulAdd
from ..cuda.utils import __has_pycuda__
from ..cuda.processing import CudaProcessing

if __has_pycuda__:
    import pycuda.gpuarray as garray


class CudaMulAdd(MulAdd):
    processing_cls = CudaProcessing

    def _init_finalize(self):
        self._init_kernel()

    def _init_kernel(self):
        self.muladd_kernel = self.processing.kernel(
            "mul_add",
            filename=get_cuda_srcfile("ElementOp.cu"),  # signature="PPiiffiiii"
        )

    def mul_add(self, dst, other, fac_dst, fac_other, dst_region=None, other_region=None):
        """
        'region' should be a tuple (slice(y_start, y_end), slice(x_start, x_end))
        """

        if dst_region is None:
            dst_coords = (0, dst.shape[1], 0, dst.shape[0])
        else:
            dst_coords = (dst_region[1].start, dst_region[1].stop, dst_region[0].start, dst_region[0].stop)
        if other_region is None:
            other_coords = (0, other.shape[1], 0, other.shape[0])
        else:
            other_coords = (other_region[1].start, other_region[1].stop, other_region[0].start, other_region[0].stop)

        delta_x = np.diff(dst_coords[:2])
        delta_y = np.diff(dst_coords[2:])
        if delta_x != np.diff(other_coords[:2]) or delta_y != np.diff(other_coords[2:]):
            raise ValueError("Invalid dst_region and other_region provided. Regions must have the same size")
        if delta_x == 0 or delta_y == 0:
            raise ValueError("delta_x or delta_y is 0")

        # can't use "int4" in pycuda ? int2 seems fine. Go figure
        # pylint: disable=E0606
        dst_x_range = np.array(dst_coords[:2], dtype=garray.vec.int2)
        dst_y_range = np.array(dst_coords[2:], dtype=garray.vec.int2)
        other_x_range = np.array(other_coords[:2], dtype=garray.vec.int2)
        other_y_range = np.array(other_coords[2:], dtype=garray.vec.int2)

        block = (32, 32, 1)
        grid = [updiv(length, b) for (length, b) in zip((delta_x[0], delta_y[0]), block)]

        self.muladd_kernel(
            dst,
            other,
            np.int32(dst.shape[1]),
            np.int32(other.shape[1]),
            np.float32(fac_dst),
            np.float32(fac_other),
            dst_x_range,
            dst_y_range,
            other_x_range,
            other_y_range,
            grid=grid,
            block=block,
        )

    __call__ = mul_add



import pytest
import numpy as np
from nabu.processing.muladd import MulAdd
from nabu.testutils import get_data
from nabu.cuda.utils import get_cuda_context, __has_pycuda__

if __has_pycuda__:
    from nabu.processing.muladd_cuda import CudaMulAdd


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.data = get_data("chelsea.npz")["data"].astype("f")  # (300, 451)
    cls.tol = 1e-7
    if __has_pycuda__:
        cls.cu_ctx = get_cuda_context(cleanup_at_exit=False)
    yield
    if __has_pycuda__:
        cls.cu_ctx.pop()


@pytest.mark.usefixtures("bootstrap")
class TestMulad:
    def test_muladd(self):
        dst = self.data.copy()
        other = self.data.copy()
        mul_add = MulAdd()

        # Test with no subregion
        mul_add(dst, other, 1, 2)
        assert np.allclose(dst, self.data * 1 + other * 2)

        # Test with x-y subregion
        dst = self.data.copy()
        mul_add(dst, other, 0.5, 1.7, (slice(10, 200), slice(15, 124)), (slice(100, 290), slice(200, 309)))
        assert np.allclose(dst[10:200, 15:124], self.data[10:200, 15:124] * 0.5 + self.data[100:290, 200:309] * 1.7)

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need Cuda/pycuda for this test")
    def test_cuda_muladd(self):
        mul_add = CudaMulAdd(ctx=self.cu_ctx)
        dst = mul_add.processing.to_device("dst", self.data)
        other = mul_add.processing.to_device("other", (self.data / 2).astype("f"))

        # Test with no subregion
        mul_add(dst, other, 3, 5)
        assert np.allclose(dst.get(), self.data * 3 + (self.data / 2) * 5)

        # Test with x-y subregion
        dst.set(self.data)
        mul_add(dst, other, 0.5, 1.7, (slice(10, 200), slice(15, 124)), (slice(100, 290), slice(200, 309)))
        assert np.allclose(
            dst.get()[10:200, 15:124], self.data[10:200, 15:124] * 0.5 + (self.data / 2)[100:290, 200:309] * 1.7
        )



from ..processing.padding_cuda import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.cuda.padding has been moved to nabu.processing.padding_cuda", do_print=True, func_name="padding_cuda"
)



import numpy as np
from ...preproc.shift_cuda import CudaVerticalShift

from ...reconstruction.sinogram_cuda import CudaSinoBuilder, CudaSinoNormalization
from ...processing.histogram_cuda import CudaPartialHistogram
from .fbp import BackprojectorHelical


try:
    from ...reconstruction.hbp import HierarchicalBackprojector  # pylint: disable=E0401,E0611

    print("Successfully imported hbp")
except:
    HierarchicalBackprojector = None

from ...cuda.utils import get_cuda_context, __has_pycuda__, __pycuda_error_msg__
from .helical_chunked_regridded import HelicalChunkedRegriddedPipeline

if __has_pycuda__:
    import pycuda.gpuarray as garray


# pylint: disable=E0606
class CudaHelicalChunkedRegriddedPipeline(HelicalChunkedRegriddedPipeline):
    """
    Cuda backend of HelicalChunkedPipeline
    """

    VerticalShiftClass = CudaVerticalShift
    SinoBuilderClass = CudaSinoBuilder
    FBPClass = BackprojectorHelical
    HBPClass = HierarchicalBackprojector

    HistogramClass = CudaPartialHistogram
    SinoNormalizationClass = CudaSinoNormalization

    def __init__(
        self,
        process_config,
        sub_region,
        logger=None,
        extra_options=None,
        phase_margin=None,
        cuda_options=None,
        reading_granularity=10,
        span_info=None,
        num_weight_radios_per_app=1000,
        diag_zpro_run=0,
    ):
        self._init_cuda(cuda_options)
        super().__init__(
            process_config,
            sub_region,
            logger=logger,
            extra_options=extra_options,
            phase_margin=phase_margin,
            reading_granularity=reading_granularity,
            span_info=span_info,
            diag_zpro_run=diag_zpro_run,
        )
        self._register_callbacks()

        self.num_weight_radios_per_app = num_weight_radios_per_app

    def _init_cuda(self, cuda_options):
        if not (__has_pycuda__):
            raise ImportError(__pycuda_error_msg__)
        cuda_options = cuda_options or {}
        self.ctx = get_cuda_context(**cuda_options)
        self._d_radios = None
        self._d_radios_weights = None
        self._d_sinos = None
        self._d_recs = None

    def _allocate_array(self, shape, dtype, name=None):
        name = name or "tmp"  # should be mandatory
        d_name = "_d_" + name
        d_arr = getattr(self, d_name, None)
        if d_arr is None:
            self.logger.debug("Allocating %s: %s" % (name, str(shape)))
            d_arr = garray.zeros(shape, dtype)
            setattr(self, d_name, d_arr)
        return d_arr

    def _process_finalize(self):
        pass

    def _post_primary_data_reduction(self, i_slice):
        self._allocate_array((self.num_weight_radios_per_app,) + self.radios_slim.shape[1:], "f", name="radios_weights")

        if self.process_config.nabu_config["reconstruction"]["angular_tolerance_steps"]:
            self.radios[:, i_slice, :][np.isnan(self.radios[:, i_slice, :])] = 0
        self.radios_slim.set(self.radios[:, i_slice, :])

    def _register_callbacks(self):
        pass

    #
    # Pipeline execution (class specialization)
    #

    def _compute_histogram(self, data=None, i_slice=None, num_slices=None):
        if self.histogram is None:
            return
        if data is None:
            data = self.recs
        my_histo = self.histogram.compute_histogram(data)
        self.histo_stack.append(my_histo)
        if i_slice == num_slices - 1:
            self.recs_histogram = self.histogram.merge_histograms(self.histo_stack)
            self.histo_stack.clear()

    def _dump_data_to_file(self, step_name, data=None):
        if data is None:
            data = self.radios
        if step_name not in self._data_dump:
            return
        if isinstance(data, garray.GPUArray):
            data = data.get()
        super()._dump_data_to_file(step_name, data=data)
